%**********************************************************
%
%   basics.sl contains functions grouped as follows:
%
%       sorting
%       occurrences (incl. functions list)
%       list buffers
%       minibuffer
%       recent files
%       shell
%       bookmarks
%
%**********************************************************

   _traceback = 1;

variable bf_list        = " /BufferList/";
variable RecentzBuf     = " /recentz/";
variable Reference_Buffer = Null_String;
variable max_r            = 32;

autoload ("display_back2menu",    "suemenu");
autoload ("ed_delete_word",       "sue");
autoload ("go_word_left",         "sue");
autoload ("go_word_right",        "sue");
autoload ("searchx",              "sue");
autoload ("replace_on_demand",    "sue");
autoload ("replace_interactivly", "sue");
autoload ("search_forward",       "sue");

%--- required prototypes ----------------------------------

define bskip_white ()           ;
define bm_keymap ()             ;
define mini_storeline (s)       ;
define occur ()                 ;
define occur_keymap()           ;
define occur_save ()            ;
define occur_shiftinit_save ()  ;
define o_srch ()                ;
define recent_switch ()         ;
define shell_exec (cmd)         ;
define shell_input ()           ;

%----------------------------------------------------------

%**********************************************************
%   sorting functions
%   sort4 is based on char stream
%   sort3 is based on rectangles
%   sort2 is a wrapper for sort4
%   sort1 is on the death row.
%**********************************************************

define sort4 (casE, descenD)
%----------------------------------------------------------
%   This sorting function needs a character stream. The
%   selection indicates which lines are included in the sort.
%
%   Sort key is an entire line as string. All lines in
%   the selection are sorted even if the selection does not
%   cover them entirely. The length of the sort key will in
%   general vary from line to line.
%
%   After sorting, the lines sorted will be highlighted
%   with the cursor on the top one of them, first column.
%
%   Case sensitive for casE==1, descending for descenD==1.
%----------------------------------------------------------
{
    variable
        i,
        n,
        begl,
        endl,
        index,
        keys;

    ifnot ( check_region (0)) return;
    %   region now in canonical form

    endl = what_line ();
    pop_mark (1);
    begl = what_line ();

    n = endl - begl + 1;
    %   count lines in region

    if(n == 1)
    {
        flush("Nothing to do: only one line to sort.");
        return;
    }

    keys = String_Type [n];

    %   extract keys and fill array

    goto_line (begl);
    for (i = 0; i < n; ++i)
    {
        bol();
        push_mark ();
        eol ();
        keys[i]  = bufsubstr();
        go_down (1);
    }

    %   Case sensitive for casE==1
    if(casE)
        index = array_sort (keys, "strcmp");
    else
        index = array_sort (keys, "strcmpi");

    %  arrange region

    goto_line (endl);
    ifnot (down (1))
    {
        eol ();
        newline ();
    }

    bol();
    push_spot ();

    for (i = 0; i < n; ++i)
    {
        if(descenD)
            goto_line (index[n-1-i] + begl);
        else
            goto_line (index[i] + begl);

        line_as_string ();
        %  goes on stack

        pop_spot ();

        insert(());
        % insert string from stack

        newline ();
        push_spot ();
    }

    pop_spot ();

    goto_line (begl);
    bol ();
    push_mark ();
    goto_line (endl + 1);
    bol ();
    del_region ();

    goto_line(endl);
    eol();
    call ("set_mark_cmd");
    go_up(n-1);
    bol();

    variable ddone = Null_String;

    if(casE) ddone = "CASE-SENSITIVE and ";
    else ddone = "case-insensitive and ";

    if (descenD==1) ddone += "descending";
    else ddone += "ascending";

    flush("Sorted: " + ddone);
}

define sort3 (casE, descenD)
%----------------------------------------------------------
%   This sorting function needs a rectangle which can just
%   be one column.
%
%   The sort key is fixed length across the rectangle.
%   Case sensitive for casE==1, descending for descenD==1.
%----------------------------------------------------------
{
    % no safety checks, called from the general sort dialog.

    ifnot ( check_rectangle ) return;
    toggle_block (1);

    variable
        thisl, thisc,
        begl, endl, begc, endc, n, index, i, keys, sortedz, tmpstr;

    %   global variables N_ROW etc. provided by check_rectangle
    endc = E_COL;
    begc = W_COL;
    endl = S_ROW;
    begl = N_ROW;

    endc++;
    %   sue's rectangles include cursor position
    %   therefore to use a sort key matching rect width
    %   we need to increase right column by 1
    %   this will also allow sorting on a single column

    n = endl - begl + 1;
    %   count lines in rectangle

    keys    = String_Type [n];
    sortedz = String_Type [n];
    %   last element of keys[] takes terminator

    %   extract keys and fill them into keys array

    thisl = what_line();
    thisc = what_column();
    goto_line (begl);
    for (i = 0; i < n; i++)
    {
        goto_column (begc);
        push_mark ();
        goto_column (endc);

        %   pop_mark (0) is implicit in bufsubstr

        keys[i] = bufsubstr ();

        go_down (1);
    }

    if (casE) index = array_sort (keys, "strcmp");
    else index = array_sort (keys, "strcmpi");

    %   get the lines into sortedz array in proper order

    for (i = 0; i < n; i++)
    {
        goto_line (index[i] + begl);
        push_mark ();
        eol ();
        tmpstr = bufsubstr ();
        sortedz[i] = tmpstr;
    }

    %   replace unsorted lines with sorted lines

    goto_line (begl);
    for (i = 0; i < n; i++)
    {
        del_to_eol ();
        if(descenD==1) insert (sortedz[n-1-i]);
        else insert(sortedz[i]);
        go_down (1);
    }

    %   restore rectangle highlight
    goto_line (thisl);
    goto_column (thisc);
    restore_rectangle ();

    variable ddone = Null_String;

    if(casE) ddone = "CASE-SENSITIVE and ";
    else ddone = "case-insensitive and ";

    if (descenD==1) ddone += "descending";
    else ddone += "ascending";

    flush("Sorted: " + ddone);
}

define sort2 ()
%----------------------------------------------------------
%   This sorting function is essentially deprecated
%   but it is used repeatedly in a couple of modules
%   and it needs to survive.
%
%   It turns now into a call to sort4 with options
%   case-insensitive and ascending set.
%----------------------------------------------------------
{
    sort4 (0,0);
    toggle_block (0);
}

define sort1 ()
%----------------------------------------------------------
%   To be replaced.
%   Used only to shape up the file listing.
%   Ascending and case-insensitive.
%----------------------------------------------------------
{
    variable
        i,
        n,
        begl,
        endl,
        begc,
        endc,
        index,
        keys;

    !if (check_region (0)) return;

    %   region now in canonical form

    endl = what_line ();
    endc = what_column ();
    pop_mark (1);
    begl = what_line ();
    begc = what_column ();

    n = endl - begl + 1;
    %   count lines in region

    %   get leftmost column of mark/cursor
    if (endc < begc)
    {
        endc;
        endc = begc;
        begc = ();
    }

    keys = String_Type [n];

    %   extract keys and fill array

    goto_line (begl);
    for (i = 0; i < n; ++i)
    {
        goto_column (begc);
        push_mark ();
        eol ();

        keys[i]  = bufsubstr();

        go_down (1);
    }

    index = array_sort (keys, "strcmpi");

    %  arrange region

    goto_line (endl);
    !if (down (1))
    {
        eol ();
        newline ();
    }

    push_spot ();

    for (i = 0; i < n; ++i)
    {
        goto_line (index[i] + begl);
        line_as_string ();
        %  on stack, we are at eol now
        pop_spot ();
        bol ();
        insert(());
        newline ();
        push_spot ();
    }

    pop_spot ();

    goto_line (begl);
    bol ();
    push_mark ();
    goto_line (endl + 1);
    bol ();
    del_region ();
    flush("Sorted ascending, case-insensitive.");
}

%**********************************************************
%
%   Occurrences
%
%   Finds all occurrences of a keyword in a buffer
%   listing them, as whole lines, in a separate buffer
%   and allowing fast switch from the list to the source.
%
%   Essentially same are the list of C or sl functions
%   and the list of shiftkey functions.
%
%**********************************************************

define occur_up ()
{
    go_up(1);

    if(is_whiteline and what_line < 6)
    {
        () = fsearch (":  ");
        recenter (window_info ('r'));
    }

    else if(is_whiteline)
        () = bsearch (":  ");

    bol();
}

define occur_down ()
{
    go_down(1);

    if(is_whiteline and what_line < 6)
    {
        () = fsearch (":  ");
    }

    else if(is_whiteline)
        () = bsearch (":  ");

    bol();
}

define occur_right ()
%----------------------------------------------------------
%   moves right on long lines in list of occurrences
%----------------------------------------------------------
{
    variable ww = window_info ('w');
    variable cc = window_info ('c');
    goto_column_best_try(cc+ww);

    % [    x    ]
    %
    %  |   |   |
    %  |   |   =    cc+ww-1
    %  |   =        what_column
    %  =            cc
}

define occur_left ()
%----------------------------------------------------------
%   moves left on long lines in list of occurrences
%----------------------------------------------------------
{
    variable cc = window_info ('c');
    goto_column_best_try (cc);
}

define occur_pagedown ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   will move up cursor one page
%----------------------------------------------------------
{
    call ("page_down");
    if(is_whiteline)
        () = bsearch(":  ");
    bol();
}

define occur_pageup ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   will move up cursor one page
%----------------------------------------------------------
{
    call ("page_up");

    if(is_whiteline or what_line < 6)
    {
        bol();
        () = fsearch(":  ");
        recenter (window_info ('r'));
    }
    bol();
}

define occur_help ()
%----------------------------------------------------------
%   ancillary function to occur ()
%----------------------------------------------------------
{
    variable
        stringA="Occur Buffer",
        stringB="    [#]",
        stringC="&Quick Help: Occurrences";

    display_help (stringA,stringB,stringC);
}

define occur_goto_line ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   go to line (as prompted for) in list of occurrences
%----------------------------------------------------------
{
    if (what_line () < 3) return;
    bol ();
    ifnot (ffind (": ")) return;

    variable listBuf = whatbuf ();
    %   store it to delete it

    Reference_Buffer = substr (whatbuf (),5,-1);
    %   necessary to handle multiple occur buffers
    %   goto_line would jump to line in wrong buffer

    ifnot (strlen(Reference_Buffer)) return;

    ifnot (bufferp (Reference_Buffer))
    {
        Reference_Buffer = Null_String;
        variable tmpbf = whatbuf ();
        set_buffer_modified_flag (0);
        sw2buf ("*scratch*");
        delbuf (tmpbf);
        flush("Reference buffer for this list not found.");
        return;
    }

    bol ();
    push_mark ();

    ifnot (ffind (": ")) return;
    %   cannot happen, used to delimit region

    variable lnstr = bufsubstr();
    lnstr = substr(lnstr,2,-1);
    variable line = integer (lnstr);
    %variable line = integer (bufsubstr);
    %   mark popped here

    sw2buf (Reference_Buffer);
    goto_line (line);
    () = ffind(LAST_SEARCH);

    delbuf(listBuf);
}

define occur_home ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   moves cursor to top of buffer
%----------------------------------------------------------
{
    bob();
    () = fsearch(":  ");
}

define occur_find ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   finds a string in list of occurrences
%----------------------------------------------------------
{
    current_mini_cmd = _function_name;
    variable sstr=Null_String;
    sstr = read__mini ("Search for:", Null_String, Null_String, 0);
    ifnot (strlen (sstr)) return;
    LAST_SEARCH=sstr;

    () = fsearch(sstr);

    %   the line and not the string is highlighted
    %   the string may occur more than once in same line
}

define occur_find_again ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   repeats last search in list of occurrences
%----------------------------------------------------------
{
    %   the line and not the string is highlighted
    %   the string may occur more than once in same line

    eol();
    if (0 == fsearch(LAST_SEARCH))
        flush("Not found.");
}

define occur_end ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   moves cursor to end of buffer, then back to last text
%----------------------------------------------------------
{
    eob();
    bskip_grey();
    bol();
}

define occur_save ()
%----------------------------------------------------------
%   special buffer is saved prompting for a file name
%----------------------------------------------------------
{
    push_spot();
    variable tmpbf = whatbuf ();
    disk_save_as ();
    rename_buffer (tmpbf);
    sw2buf(tmpbf);
    occur_keymap();
    pop_spot();
}

define occur_quit ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   delete buffer with list of occurrences
%----------------------------------------------------------
{
    delbuf (whatbuf);
    sw2buf (Reference_Buffer);
    Reference_Buffer=Null_String;
}

define occur_stub (str)
%----------------------------------------------------------
%   ancillary function to occur ()
%   stub for list of occurrences
%----------------------------------------------------------
{
    sw2buf (o_ListBuffer);
    erase_buffer ();
    newline ();
    loop (8) insert_char (' ');
    insert
    ("Press Enter on a line to go to that line in reference buffer.");
    newline ();
    newline ();
}

define occur (sstr)
%----------------------------------------------------------
%   This function does the real work finding and listing
%   occurrences of the argument.
%
%   SEARCH_TYPE will determine simple or regex search.
%----------------------------------------------------------
{
    %   function will run when selection present,
    %   selection wil be lost,
    %   occurrences on selection lines wil be tagged

    variable fromL=0, toL=0;
    variable ln_start = what_line();

    if (markp == 1) toggle_block(0);

    if(what_block==1)
    {
        if(check_rectangle)
        {
            toL =   S_ROW;
            fromL = N_ROW;
        }
    }

    else if(check_region(0))
    {
        toL = what_line();
        call("exchange");
        fromL = what_line();
    }

    toggle_block(0);

    Reference_Buffer = whatbuf ();

    variable
        lnstr      = Null_String,
        ln_ntmp    = 0,
        ln         = 0;

    push_spot();

    %   give list a name dependent on source
    o_ListBuffer =  "/O/ " + Reference_Buffer;

    %   get rid of any previous list
    delbuf (o_ListBuffer);
    sw2buf (o_ListBuffer);

    %   limit literal sstr to half screen width
    %   regex pattern are mostly shorter
    ifnot(SEARCH_TYPE)
        sstr = substr (sstr, 1, SCREEN_WIDTH/2);

    occur_stub (sstr);
    LAST_SEARCH = sstr;

    sw2buf (Reference_Buffer);
    bob();

    variable sf;
    if (SEARCH_TYPE) sf = &re_fsearch;
    else sf = &fsearch;

    while (@sf(sstr))
    {
        ln = what_line ();

        if (ln > 99999)
        {
            flush("Capacity exceeded, only partial results.");
            sleep(2);
            break;
        }

        bol();
        skip_white();

        push_mark ();
        eol ();
        bufsubstr ();

        sw2buf (o_ListBuffer);

        if(ln >= fromL and ln <= toL)
            vinsert ("*%5d:  ", ln);
        else
            vinsert (" %5d:  ", ln);

        %   insert from stack
        insert (());
        newline ();

        sw2buf (Reference_Buffer);
    }

    sw2buf (Reference_Buffer);

    pop_spot();
    %   cursor in original position in reference buffer

    ifnot(ln)
    {
        delbuf(o_ListBuffer);
        sw2buf(Reference_Buffer);
        flush("Nothing found.");
        return;
    }

    %   initialize list
    sw2buf (o_ListBuffer);
    bob ();

    %   position cursor in list, first match in selection
    %   if none, according to cursor position in reference buffer

    variable rr;
    rr = bol_fsearch ("*");

    if(rr);
    %   just stay there

    else
    {
        () = fsearch(":  ");
        do
        {
            bol();
            push_mark();
            go_right(6);
            lnstr = bufsubstr();
            ln_ntmp = integer(lnstr);
            if (ln_ntmp >= ln_start) break;
        }
        while (down(1));
        if (ln_ntmp != ln_start) go_up(1);
    }

    if(is_whiteline) go_up(1);
    bol();

    occur_keymap ();
    set_buffer_modified_flag (0);
    set_readonly (1);

    set_status_line
    ("  %p  |  Reference Buffer: "+Reference_Buffer+"  |  F1=Help   ", 0);

    %   use menu bar as title
    variable m = "&Occurrences:  " + sstr;
    dummy_menu (m);

    flush("Press Enter to go to current entry in reference buffer.");
    %   useful in long lists where top is not seen
}

define occur_sl_functions ()
%----------------------------------------------------------
%   Will list sl functions heuristically.
%   Results are in general ok but sometimes not exhaustive.
%   False positives will be avoided if comments do not contain
%   round brackets.
%----------------------------------------------------------
{
    %   no selection allowed (since it may get extended)
    if (markp)
    {
        flush(s_reminder);
        return;
    }

    Reference_Buffer = whatbuf ();

    variable
        ln_near  = -1,
        ln_ntmp  = -1,
        ln       = -1,
        thisdir  = Null_String,
        thisfile = Null_String,
        thisfil1 = Null_String,
        lnstr    = Null_String,
        ccmnd    = Null_String;

    (thisfile, thisdir,,) = getbuf_info ();

    ifnot (strlen(thisfile))
    {
        flush("Save buffer first.");
        return;
    }

    ln_near = what_line ();
    push_spot ();
    %   we want the cursor to stay where it is

    thisfile = thisdir + thisfile;
    thisfil1 = thisfile + ".tmp1";

    %   give the function list a name dependent on source
    f_ListBuffer =  "/F/ " + whatbuf ();
    sw2buf(f_ListBuffer);
    erase_buffer();

    setbuf (Reference_Buffer);
    bob();

    % we do not want to pick the following  d e f i n e

    variable defkeyword = "d" + "efine ";
    % we do not want to pick some lines when searching for the keyword

    while (fsearch(defkeyword))
    {
        if(bfind(char(37)))
        {
            % char 27 is slang comment character
            eol();
            continue;
        }

        ifnot(bolp)
        {
            () = bfind(" ");
            ifnot (looking_at(" define ")) continue;
        }

        eol();
        if(bfind(";")) continue;
        %   prototype skipped

        bol();
        %   skip the line beginning this while-loop
        if(ffind("\"))")) continue;

        %   entire line to the stack
        bol();
        push_mark ();
        eol();
        bufsubstr();

        ln = what_line();
        setbuf (f_ListBuffer);
        if (ln > 99999)
        {
            flush("Line number too big, exiting ...");
            break;
        }
        else vinsert ("%5d:  ", ln);

        %   insert from the stack
        insert (());
        newline ();
        setbuf (Reference_Buffer);
    }

    pop_spot();
    %   cursor back to original position in reference buffer

    setbuf (f_ListBuffer);
    bob();
    if (eobp==bobp)
    {
        delbuf(whatbuf);
        flush("Not found.");
        return;
    }

    newline ();
    loop (8) insert_char (' ');
    insert
    ("Press Enter on a line to go to that line in reference buffer.");
    newline ();
    newline ();
    () = fsearch(":  ");

    %   position cursor in list
    %   according to cursor position in reference buffer
    do
    {
        bol();
        push_mark();
        go_right(5);
        lnstr = strtrim(bufsubstr());
        %   mark is popped

        ln_ntmp = integer(lnstr);

        if (ln_ntmp < ln_near) continue;
        break;
    }
    while (down(1));
    bol();

    occur_keymap ();
    set_buffer_modified_flag (0);
    set_readonly (1);

    set_status_line
    ("  %p  |  Functions in: "+Reference_Buffer+"  |  F1=Help", 0);

    variable m = "&Functions";
    dummy_menu(m);
}

define occur_c_functions ()
{
    variable
        ln_near  = what_line (),
        ln_ntmp  = -1,
        interbf  = Null_String,
        interbfd = Null_String,
        thisdir  = Null_String,
        thisfile = Null_String,
        thisfil1 = Null_String,
        lnstr    = Null_String,
        ccmnd    = Null_String;

    (thisfile, thisdir,,) = getbuf_info ();

    if (buffer_modified)
    {
        interbf  = "tmp__" + thisfile;
        interbfd = thisdir + interbf;

        sw2buf(interbf);
        erase_buffer(interbf);
        insbuf(Reference_Buffer);

        variable written = write_buffer(interbfd);

        if(written < 0)
        {
            delbuf(interbf);
            () = delete_file (interbfd);
            sw2buf(Reference_Buffer);
            goto_line(ln_near);
            return;
        }

        %   alternative thisfile
        thisfile = interbf;
        delbuf(interbf);
        sw2buf(Reference_Buffer);
        goto_line(ln_near);
    }

    thisfile = thisdir + thisfile;
    thisfil1 = thisfile + ".tmp1";

    %   give the function list a name dependent on source
    f_ListBuffer =  "/F/ " + Reference_Buffer;
    sw2buf(f_ListBuffer);

    %   let ctags do the chores
    %   -x  output is a table (or so they say)
    %   -n  procedure relies only on line number, quicker
    %   -u  no sort on name, result will be sorted on line number
    %   --c-kinds=f     consider only c-functions (no variables, macros etc.)

    ccmnd = "ctags -x -n -u --c-kinds=f " + thisfile +  "|grep ' function ' > " + thisfil1;

    try
    {
        () = run_shell_cmd (ccmnd);
    }
    catch AnyError:
    {
        flush("Unable to run ctags.");
        delbuf(whatbuf);
        return;
    }

    erase_buffer();
    () = insert_file (thisfil1);
    () = delete_file (thisfil1);
    () = delete_file (interbfd);
    bob ();
    replace (thisfile, "");

    %   clean up the list
    while (fsearch(" function "))
    {
        skip_grey();
        skip_black();
        del_to_bol();
    }

    %   some formatting
    bob();

    while(re_fsearch("^[ \t]*$"))
    {
        if (eobp) break;
        delete_line();
    }

    if (eobp==bobp)
    {
        delbuf(whatbuf);
        flush("Not found.");
        return;
    }

    bob();
    do
    {
        %   get the line number into a string
        trim();
        push_mark();
        () = ffind(" ");
        lnstr = strtrim(bufsubstr());
        %   mark is popped

        %   delete the line number
        bol();
        ed_delete_word();

        variable ln_n = 0;
        try
        {
            ln_n = integer(lnstr);
        }
        catch AnyError:
        {
            ln_n = 0;
        }

        %   insert formatted line number
        if (ln_n > 99999)
        {
            flush("Line number too big, exiting ...");
            break;
        }
        else ifnot (ln_n) del_to_eol();
        % do nothing

        else if (strlen(lnstr)) vinsert ("%5d:  ", integer(lnstr));
    }
    while (down(1));

    %   list is already sorted by line number

    bob();
    newline ();
    loop (8) insert_char (' ');
    insert
    ("Press Enter on a line to go to that line in reference buffer.");
    newline ();
    newline ();
    () = fsearch(":  ");

    %   position cursor in list
    %   according to cursor position in reference buffer
    do
    {
        bol();
        push_mark();
        go_right(5);
        lnstr = strtrim(bufsubstr());
        %   mark is popped

        ln_ntmp = integer(lnstr);

        if (ln_ntmp < ln_near) continue;
        break;
    }
    while (down(1));

    if(is_whiteline) go_up(1);
    bol();

    occur_keymap ();
    set_buffer_modified_flag (0);
    set_readonly (1);

    set_status_line
    ("  %p  |  Functions in: "+Reference_Buffer+"  |  F1=Help", 0);

    variable m = "&Functions";
    dummy_menu(m);
    flush(" ");
}

define occur_functions ()
%----------------------------------------------------------
%   Will list C functions using the utility ctags or
%   routing to another function for SLang. It can be made
%   to work for all languages covered by ctags but it works
%   using disk files. Therefore, if the current buffer is
%   not saved there might be discrepancies since temp files
%   are used.
%
%   Unfortunately, ctags may produce slightly differing
%   output from version to version and from distro to
%   distro. if the output is not a clean list, the code
%   below must be revised, we expect an output line like
%   short_name line_type line_number file_name full_function_name
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    Reference_Buffer = whatbuf ();
    variable ndx = 0;

    ndx = is_substr (Reference_Buffer, ".sl");
    if(ndx > 1)
    {
        occur_sl_functions();
        return;
    }

    ndx = is_substr (Reference_Buffer, ".c");
    if(ndx > 1)
    {
        %   possibly c file
        occur_c_functions();
        return;
    }

    ndx = is_substr (Reference_Buffer, ".h");
    if(ndx > 1)
    {
        %   possibly header file
        occur_c_functions();
        return;
    }

    flush("Extension 'c' or 'sl' required for this feature");
    sleep(1);
}

define occur_shiftkeys_show ()
%----------------------------------------------------------
%   ancillary to occur_shiftkeys ()
%----------------------------------------------------------
{
    %   only proceed on valid lines
    bol ();
    ifnot (ffind(":  ")) return;

    %   buffers involved
    variable ttmpbf = Null_String;
    ttmpbf = whatbuf ();
    variable tmp_bf = " /shiftkey_details/";

    %   get the line number from begin of line
    bol ();
    push_mark ();
    ifnot (ffind(":  ")) return;
    push_spot();
    variable lline = integer (bufsubstr ());
    %   mark popped here

    %   get the file name

    set_readonly(0);
    goto_column(9);
    set_readonly(1);

    push_mark();
    skip_black ();
    variable ffile = bufsubstr();
    ffile = strcat (JED_ROOT+"/lib/",ffile);

    %   load file, find line, show it
    sw2buf (tmp_bf);
    erase_buffer();

    insert_file (ffile);
    goto_line (lline);
    () = bsearch(char(100)+"efine ");
    newline();
    newline();
    insert("line ");
    insert(string(lline));
    insert(",  ");
    insert(ffile);
    newline();
    newline();
    () = bsearch("line ");
    recenter (2);
    () = fsearch("Shift pressed:");
    hide_status_indicators(1);
    set_status_line("  Press any key to exit",0);
    variable m = "&Shiftkeys Details";
    dummy_menu(m);

    update (1);

    %   wait for a key to close
    flush_input(getkey ());

    sw2buf(ttmpbf);
}

define occur_back2menu ()
{
    delbuf(whatbuf);
    menu_last_entry();
}

define occur_shiftkeymap ()
{
    reset_slim_keymap ();
    $1 = "shiftkeys_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i<256; i++)
        undefinekey (char(i), $1);

    definekey ("occur_down",            Key_Down,   $1);
    definekey ("occur_up",              Key_Up,     $1);
    definekey ("occur_shiftkeys_show",  Key_Enter,  $1);
    definekey ("occur_back2menu",       "^W",       $1);
    definekey ("occur_back2menu",       Key_Escape, $1);
    definekey ("occur_shiftinit_save",  "^S",       $1);
    definekey ("display_back2menu",     Key_F10,    $1);
}

define occur_shiftinit_save ()
%----------------------------------------------------------
%   special buffer is saved prompting for a file name
%----------------------------------------------------------
{
    push_spot();
    variable tmpbf = whatbuf ();
    disk_save_as ();
    rename_buffer (tmpbf);
    sw2buf(tmpbf);
    occur_shiftkeymap ();
    pop_spot();
}

define occur_shiftkeys ()
%----------------------------------------------------------
%   list of shift keys in use but not in a keymap
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    %   open temporary buffer
    variable b_tmpbf = " /Shiftkey Commands/";
    variable cmdstr = Null_String;
    variable fmask = strcat (JED_ROOT,"/lib/*.sl");
    delbuf(b_tmpbf);
    sw2buf(b_tmpbf);

    cmdstr = strcat
    ("grep -ins 'Shift pressed:' ", fmask, " 2> ~/", string(random(-1,0)), ".tmp &");

    try
    {
        () = run_shell_cmd (cmdstr);
        %   output goes into current buffer
    }
    catch AnyError:
    {
        flush("Unable to run grep.");
        delbuf(whatbuf);
        return;
    }

    bob();

    while (fsearch("Shift pressed:"))
        replace_chars (strlen("Shift pressed:"),"");
    bob();
    while (fsearch(":")) replace_chars (1," ");
    bob();

    variable i=0, j=0, lastln=16, ln=0;
    variable bfilez = String_Type [lastln];
    variable ffilez = String_Type [lastln];
    variable llinez = String_Type [lastln];
    variable hhintz = String_Type [lastln];

    bfilez [*] = Null_String;
    ffilez [*] = Null_String;
    llinez [*] = Null_String;
    hhintz [*] = Null_String;

    %   load hits into array for further operations
    while (fsearch(".sl"))
    {
        if (i > lastln-1) break;

        bol();
        ifnot(ffind("%   "))
        {
            del_to_eol();
            continue;
            %   irrelevant line
        }
        bol();

        %   file name

        push_mark ();
        skip_black ();
        ffilez[i] = bufsubstr ();
        bfilez[i] = path_basename(ffilez[i]);

        %   line number

        skip_grey ();
        push_mark();
        skip_black ();
        llinez[i] = bufsubstr();

        %   explanatory text

        () = ffind("%");
        go_right(1);
        skip_grey ();
        push_mark();
        eol();
        hhintz[i] = bufsubstr();

        i++;
    }
    if (i < lastln) lastln=i;

    ifnot (i)
    {
        flush("List is empty");
        delbuf(whatbuf);
        return;
    }

    erase_buffer ();

    %   header text for list
    newline ();
    goto_column(9);
    variable instruct = "Press Enter on a line to see details.";
    insert (instruct);
    newline ();
    newline ();

    for (i=0; i<lastln; i++)
    {
        ln = integer(llinez[i]);

        if (ln > 99999)
        {
            flush("Line number too big, skipping ...");
            sleep(1);
            continue;
        }
        else vinsert ("%5s:  ",llinez[i]);

        goto_column(9);
        insert (bfilez[i]);
        goto_column(34);
        insert (hhintz[i]);
        eol();
        newline();
    }
    bob();
    () = fsearch (":  ");

    %   prepare for the list
    set_readonly(1);
    set_buffer_modified_flag (0);
    set_status_line
    ("  Escape: exit  |  Enter: details  |  Up, Down: navigate  |  ctrl-s: save", 0);
    variable m = "&Shiftkey Commands (not in keymaps)";
    dummy_menu(m);

    occur_shiftkeymap();
}

define occur_simple ()
%----------------------------------------------------------
%   list occurrences case-insensitive, simple search
%----------------------------------------------------------
{
    current_mini_cmd = _function_name;
    variable sstr     = Null_String;
    variable tmpbf   = whatbuf ();

    %   use block on a single line as key, if any
    if(1 == chk_stream) sstr = bufsubstr ();
    else if(markp==1)
    {
        %   get string from rectangle as well
        %   viewer has no rectangles
        copy_rect();

        sw2buf("*scratch*");
        eob();
        newline();
        push_spot();
        call("yank");
        pop_spot();
        push_mark();
        eol();
        sstr = bufsubstr ();
        sw2buf(tmpbf);
    }

    else if(markp)
    {
        o_srch();
        return;
    }

    toggle_block (0);

    variable o_prompt =
    "Non-regex case-insensitive occurrences of:";

    ifnot(strlen(sstr))
        sstr = read__mini (o_prompt, Null_String, Null_String, 0);

    ifnot (strlen(sstr)) return;

    if(strlen(sstr) > 3) mini_storeline (sstr);

    o_ListBuffer = Null_String;
    %   get rid of any leftovers

    LAST_SEARCH = sstr;
    SEARCH_TYPE = 0;
    occur (sstr);
}

define occur_keymap ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   will initialize keymap for list of occurrences
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "occur_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<127; i++)
        undefinekey (char(i), $1);

    definekey ("occur_find",        "^F",       $1);
    definekey ("occur_save",        "^S",       $1);
    definekey ("occur_test",        "^T",       $1);
    definekey ("occur_quit",        "^W",       $1);
    definekey ("occur_down",        Key_Down,   $1);
    definekey ("occur_end",         Key_End,    $1);
    definekey ("occur_goto_line",   Key_Enter,  $1);
    definekey ("occur_quit",        Key_Escape, $1);
    definekey ("occur_help",        Key_F1,     $1);
    definekey ("occur_find_again",  Key_F3,     $1);
    definekey ("occur_home",        Key_Home,   $1);
    definekey ("occur_left",        Key_Left,   $1);
    definekey ("occur_pagedown",    Key_PgDn,   $1);
    definekey ("occur_pageup",      Key_PgUp,   $1);
    definekey ("occur_right",       Key_Right,  $1);
    definekey ("occur_up",          Key_Up,     $1);
}

%**********************************************************
%   Functions to list buffers and switch buffer
%
%   dependencies:
%   bufed_start
%     bufed_list
%     bufed_keymap
%       bufed_sw2buf
%         bufed_get
%
%**********************************************************

define bufed_up ()
{
    go_up(1);

    if(is_whiteline and what_line < 6)
    {
        () = fsearch (slash);
        recenter (window_info ('r'));
    }

    else if(is_whiteline)
        () = bsearch (slash);

    bol();
}

define bufed_down ()
{
    go_down(1);

    if(is_whiteline and what_line < 6)
    {
        () = fsearch (slash);
    }

    else if(is_whiteline)
        () = bsearch (slash);

    bol();
}

define bufed_list ()
%----------------------------------------------------------
%   Will list buffers accessible to user
%----------------------------------------------------------
{
    check_buffers ();

    variable
        name,
        flags,
        name_col = 5,
        dir_col  = 29;

    delbuf(bf_list);
    sw2buf(bf_list);

    newline ();
    goto_column (name_col);
    insert ("Buffer");
    goto_column(dir_col);
    insert("Directory and File");
    newline ();
    newline ();

    loop (buffer_list())
    {
        name = ();
        if (is_substr(name,slash)) continue;
        %   sue's special buffers not listed

        if(substr(name,1,1) == " ") continue;
        %   jed's special buffers not listed

        flags = getbuf_info (name);
        %   will push plenty into stack
        %   not easily readable but quick

        goto_column (name_col);

        %   buffer name in quotes to evidence whitespace
        insert_char ('"');

        insert(());
        %   buffer name inserted

        insert_char ('"');

        %   what to do with excessive length?
        variable lc = what_column();
        if (lc < dir_col - 1) goto_column(dir_col);
        else goto_column(lc+5);

        insert(());
        insert(());
        %   dir/file inserted

        newline ();
    }

    if(viewer_only)
    {
        bob();
        () = fsearch("*scratch*");
        delete_line ();

        %   anything left?
        bol();
        ifnot(fsearch(slash)) delbuf(whatbuf);
    }
}

define bufed_get ()
%----------------------------------------------------------
%   Will get the buffer name associated
%   with the current line in buffers list
%----------------------------------------------------------
{
    EXIT_BLOCK  pop_spot ();

    variable buf;

    push_spot ();
    bol ();

    ifnot (ffind_char ('"')) return Null_String;

    go_right (1);
    push_mark ();
    ifnot (ffind_char ('"'))
    {
        pop_mark (1);
        return Null_String;
    }

    buf = bufsubstr ();
    ifnot (bufferp (buf)) buf = "";
    return buf;
}

define bufed_sw2buf ()
%----------------------------------------------------------
%   Will switch to the buffer associated
%   with the current line in buffers list
%----------------------------------------------------------
{
    variable buf = Null_String;

    buf = bufed_get ();

    ifnot(strlen(buf)) return;

    delbuf (bf_list);

    sw2buf (buf);
}

define bufed_keymap ()
%----------------------------------------------------------
%   Creates a keymap for the buffers list
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "bufed_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i<256; i++)
        undefinekey (char(i), $1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Home,  $1);
    undefinekey (Key_End,   $1);
    undefinekey (Key_Del,   $1);

    definekey ("bufed_up",      Key_Up,     $1);
    definekey ("bufed_down",    Key_Down,   $1);
    definekey ("bufed_sw2buf",  Key_Enter,  $1);
    definekey ("slim_quit",     "^W",       $1);
    definekey ("slim_quit",     Key_Escape, $1);
}

define bufed_start ()
%----------------------------------------------------------
%   Initialize the buffers list
%----------------------------------------------------------
{

    if(nwinz) return;
    %   split screen

    variable this_buf = Null_String;
    this_buf = slash + whatbuf () +slash;

    bufed_list ();
    bob();
    () = fsearch (this_buf);

    bufed_keymap ();

    set_status_line
    ("    Escape: exit    |    Up, Down: navigate    |    Enter: switch to buffer" ,0);

    variable m = "&Buffer List";
    dummy_menu(m);

    bob();
    () = fsearch(slash);

    set_readonly (1);
}

%**********************************************************
%   Functions for MINIBUFFER operations
%   incl. recall of previous commands
%   trim and only save commands of at least 3 characters
%   save only commands different from old ones in store
%   separate history stores are used
%       0   for search strings
%       1   for paths and files
%       2   for motley mix
%   They are determined by the variable mS from site.sl
%**********************************************************

%   Mini_Top is the max. number of lines allowed for history
%   Mini_History is a string array for history lines
%       distinct for sundry and files
%   Mini_Last is the index of line reached when recalling history

variable
    Mini_Top = 16,
    Mini_Stores = 3,
    Mini_History = String_Type [Mini_Stores, Mini_Top],
    Mini_Last = Integer_Type [Mini_Stores];

variable heml = 3;
%history entry minimum length

Mini_Last[*] = 0;
Mini_History[*,*] = Null_String;

define mini_test ()
{

}

define mini_next ()
%----------------------------------------------------------
%   Get next command from minibuffer's history
%   'next' means decreased index, cycle around at bottom
%----------------------------------------------------------
{
    erase_buffer ();

    variable ndx = Mini_Last[mS]-1;
    if (ndx < 0) ndx += Mini_Top;

    while (strlen(Mini_History[mS, ndx]) < heml and ndx > 0)
        ndx--;

    if (strlen(Mini_History[mS, ndx]) > heml-1)
        insert (Mini_History[mS,ndx]);

    Mini_Last[mS] = ndx;
}

define mini_prev ()
%----------------------------------------------------------
%   Get previous command from minibuffer's history
%   'previous' means increased index, cycle around at top
%----------------------------------------------------------
{
    erase_buffer ();

    if (strlen(Mini_History[mS, Mini_Last[mS]]) < heml)
    {
        Mini_Last[mS]=0;

        if (strlen(Mini_History[mS, Mini_Last[mS]]) < heml)
            return;
    }

    insert (Mini_History[mS, Mini_Last[mS]]);
    Mini_Last[mS]++;
    Mini_Last[mS] = Mini_Last[mS] mod Mini_Top;
}

define mini_storeline (s)
%----------------------------------------------------------
%   Will copy accepted line to history store
%----------------------------------------------------------
{
    %   s is already strtrim(s)
    %   there is no whitespace at both ends, already trimmed

    %   keep short lines out of history
    if (strlen(s) < heml) return;

    %   insert argument s at begin of history,
    %   increase index for older entries discarding top
    %   a temp array is used for operations

    variable i=0, j=0, strchk=0;
    variable Mini_Tmp = String_Type [Mini_Stores, Mini_Top];
    Mini_Tmp [*,*] = Null_String;

    Mini_Tmp [mS,j] = s;
    j++;
    %   slang error if index incremented in array!!!!!

    %   we copy to temp array history entries not empty
    %   and not overlapping with argument s
    %   j starts with 1 because index 0 is taking fresh entry

    for (i=0; i<Mini_Top; i++)
    {
        if (j==Mini_Top) break;

        if(strlen(Mini_History[mS,i]) < heml)
            continue;

        strchk = is_substr(s, Mini_History[mS,i]);
        %   both arguments have positive length

        %   if entry has no overlap with history line, keep it
        ifnot (strchk)
        {
            Mini_Tmp [mS, j] = Mini_History[mS,i];
            j++;
        }
    }

    %   array Mini_Tmp has now the refreshed history
    %   no repetitions, no short length strings

    Mini_History[mS, *] = Mini_Tmp[mS, *];
}

define mini_remove ()
%----------------------------------------------------------
%   Will remove from history the currently displayed line
%----------------------------------------------------------
{
    variable i=0, c_line = Null_String;
    c_line = strtrim(line_as_string);
    erase_buffer ();

    ifnot(Mini_Last[mS]) return;
    %   line just typed, not in history

    if(strlen(c_line) < heml) return;

    mini_storeline (c_line);

    %   that line is now at begin of history array
    %   we discard it

    for (i=0; i < Mini_Top-1; i++)
        Mini_History[mS,i] = Mini_History[mS,i+1];

    Mini_Last[mS] = 0;
}

define mini_quit ()
%----------------------------------------------------------
%   Will quit minibuffer passing on nothing
%----------------------------------------------------------
{
    variable
        m_complete = "*Completions*";

    ifnot(_stkdepth) "*";

    if(bufferp(m_complete)) delbuf(m_complete);

    call ("exit_mini");
    Mini_Last[mS] = 0;
}

define mini_exit ()
%----------------------------------------------------------
%   Will exit the minibuffer passing on the current line.
%----------------------------------------------------------
{
    variable c_line = Null_String;
    c_line = strtrim(line_as_string);
    mini_storeline (c_line);

    variable
        m_complete = "*Completions*";
    if(bufferp(m_complete)) delbuf(m_complete);

    call ("exit_mini");
    Mini_Last[mS] = 0;
}

define mini_paste ()
%----------------------------------------------------------
%   Will paste from the clipboard in the minibuffer
%----------------------------------------------------------
{
    call("yank");
}

define mini_copy ()
%----------------------------------------------------------
%   Will copy entire minibuffer to clipboard
%----------------------------------------------------------
{
    if (eobp==bobp) return;

    push_spot ();
    bob();
    push_mark ();
    eob ();
    call ("copy_region");
    % mark gone
    pop_spot ();
}

define mini_left_key ()
%----------------------------------------------------------
%   Go left one character or one word
%----------------------------------------------------------
{
    if(CTRLKEY_PRESSED) go_word_left();
    else call ("previous_char_cmd");
}

define mini_right_key ()
%----------------------------------------------------------
%   Go right one character or one word
%----------------------------------------------------------
{
    if(CTRLKEY_PRESSED) go_word_right();
    else call ("next_char_cmd");
}

define mini_del_bol ()
%----------------------------------------------------------
%   Will delete to begin of line in the minibuffer
%----------------------------------------------------------
{
    push_mark ();
    bol ();
    del_region();
}

define mini_del_eol ()
%----------------------------------------------------------
%   Will delete to end of line in the minibuffer
%----------------------------------------------------------
{
    push_mark ();
    eol ();
    del_region();
}

define mini_del_key ()
%----------------------------------------------------------
%   Will delete current character - or region left or right
%----------------------------------------------------------
{
    if(ALTKEY_PRESSED) mini_del_bol();

    else if(CTRLKEY_PRESSED) mini_del_eol();

    else del();
}

define mini_BS_key ()
%----------------------------------------------------------
%   Will delete left character or current word
%----------------------------------------------------------
{
    if(CTRLKEY_PRESSED) ed_delete_word();

    else call ("backward_delete_char");
}

define mini_cmd_help ()
{
    %   current_mini_cmd is set in the function
    %   that is using the minibuffer

    variable
        stringA = "MiniBuffer",
        stringB = current_mini_cmd,
        stringC = "&MiniBuffer Current Command";

    ifnot(strlen(stringB)) return;

    otherwindow();
    variable tmpbf = whatbuf;

    display_help(stringA,stringB,stringC);

    sw2buf(tmpbf);
    otherwindow();
    %   avoid trouble
}

define mini_quickhelp ()
%----------------------------------------------------------
%   Shift not pressed:  generic minibuffer help
%   Shift pressed:      help on function currently executing
%----------------------------------------------------------
{
    if(SHIFTKEY_PRESSED)
    {
        mini_cmd_help();
        return;
    }

    variable
        stringA = "MiniBuffer",
        stringB = " ",
        stringC = "&MiniBuffer Keybindings";

    otherwindow();

    display_help(stringA,stringB,stringC);

    otherwindow();
    %   avoid trouble
}

define mini_insert_key ()
{
    if(ALTKEY_PRESSED)
        insert_code_for_key ();
}

define mini_cmplt_down ()
{
    variable
        m_complete = "*Completions*",
        tmpbf = Null_String;

    if (MINIBUFFER_ACTIVE) tmpbf = whatbuf();
    else return;

    if(bufferp(m_complete))
    {
        setbuf(m_complete);
        set_status_line ("    PageUp/PageDown to page through lengthy list", 0);
        hide_status_indicators(1);
        call("page_up");
        sw2buf(tmpbf);
    }
}

define mini_cmplt_up ()
{
    variable
        m_complete = "*Completions*",
        tmpbf = Null_String;

    if (MINIBUFFER_ACTIVE) tmpbf = whatbuf();
    else return;

    if(bufferp(m_complete))
    {
        setbuf(m_complete);
        set_status_line ("    PageUp/PageDown to page through lengthy list", 0);
        hide_status_indicators(1);
        call("page_down");
        sw2buf(tmpbf);
    }
}

define mini_keymap ()
%----------------------------------------------------------
%   Will set up the minibuffer,
%   called only once from site.sl
%----------------------------------------------------------
{
    variable i;
    $1 = "Mini_slim_Map";
    %   must be same name as in ledit.c

    %   reset_slim_keymap: no, function is used
    %   exactly once when the editor starts

    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");

    %   use_keymap: no, already done in ledit.c

    %definekey ("mini_test",         "^T",       $1);

    definekey ("mini_copy",         "^C",       $1);
    definekey ("mini_paste",        "^V",       $1);
    definekey ("mini_quit",         "^W",       $1);
    definekey ("mini_quit",         Key_Escape, $1);
    definekey ("mini_remove",       "^Y",       $1);
    definekey ("mini_BS_key",       Key_BS,     $1);
    definekey ("mini_del_key",      Key_Del,    $1);
    definekey ("mini_next",         Key_Down,   $1);
    definekey ("eob",               Key_End,    $1);
    definekey ("mini_exit",         Key_Enter,  $1);
    definekey ("mini_quickhelp",    Key_F1,     $1);
    definekey ("bob",               Key_Home,   $1);
    definekey ("mini_insert_key",   Key_Ins,    $1);
    definekey ("mini_left_key",     Key_Left,   $1);
    definekey ("mini_right_key",    Key_Right,  $1);
    definekey ("mini_prev",         Key_Up,     $1);
    definekey ("mini_complete",     Key_Tab,    $1);
    definekey ("mini_complete",     Key_Space,  $1);
    definekey ("mini_cmplt_up",     Key_PgDn,   $1);
    definekey ("mini_cmplt_down",   Key_PgUp,   $1);

    %   also available
    %   kbd_quit  Control_backslash  from keymap.c
}

%**********************************************************
%   Recent Files
%   A list of recently edited files is kept in
%   the user's home directory.
%   The list can be browsed to open files again.
%   File names are appended to it when files are saved.
%**********************************************************

define recent_quit ()
%----------------------------------------------------------
%   Close the recent files list with a key shortcut
%   directly from the list itself.
%----------------------------------------------------------
{
    delbuf (whatbuf);
}

define recent_remove_line ()
%----------------------------------------------------------
%   will remove a line from the recent list displayed
%----------------------------------------------------------
{
    EXIT_BLOCK set_readonly (1);

    variable fstrng = RecentFiles;

    %   this is from the dummy menu
    if (globl_str=="&Recently Accessed Directories")
        fstrng = RecentDirs;

    bol();
    ifnot (ffind(slash)) return;
    %   no entry to delete on this line

    set_readonly (0);
    bol();
    del_to_eol ();
    del();

    bob();
    %   any entry left?
    if (fsearch(slash))
    {
        %   save region
        push_spot();
        push_mark();
        eob();
        () = write_region_to_file (fstrng);
        pop_spot();
    }

    else recent_quit();
}

define recent_open_file ()
%----------------------------------------------------------
%   Shift not pressed:  will open a file from the list
%                       of recently edited files
%
%   Shift pressed:      will open only path from the list
%                       of recently edited files
%----------------------------------------------------------
{
    bol();
    ifnot (ffind(slash)) return;

    variable recentln = Null_String;

    bol ();
    push_mark();
    eol();
    recentln = strtrim_beg(bufsubstr());

    ifnot (strlen(recentln)) return;
    if (file_status (recentln) != 1)
    {
        flush("Not found.");
        return;
    }

    recent_quit ();
    %   otherwise RecentzBuf survives file opening

    %   press shift to open only the path of highlighted file
    if(SHIFTKEY_PRESSED)
    {
        variable recentdir = Null_String;
        recentdir = path_dirname (recentln);
        dls (recentdir);
    }

    else () = find_file (recentln);
}

define recent_open_dir ()
%----------------------------------------------------------
%   open a directory from the list of recently accessed
%----------------------------------------------------------
{
    variable recentln = Null_String;

    bol ();
    ifnot (ffind(slash)) return;

    push_mark();
    eol();
    recentln = bufsubstr();

    if (file_status (recentln) != 2)
    {
        flush("Not found.");
        return;
    }

    recent_quit ();
    dls (recentln);
}

define recent_load ()
%----------------------------------------------------------
%   open file or directory from list
%----------------------------------------------------------
{
    variable m = globl_str;
    globl_str = Null_String;

    if (m == "&Recently Accessed Directories")
    {
        recent_open_dir ();
        return;
    }

    recent_open_file ();
}

define recent__files_update ()
%----------------------------------------------------------
%   Will pre-pend the current file name to the list of
%   recently edited files (i.e the disk file) when saving.
%   This function is only used as a hook after saving,
%   files opened but not saved will not be considered.
%----------------------------------------------------------
{
    variable fst = file_status (RecentFiles);
    if (fst!=0 and fst!=1) return;

    variable
        llen      = 0,
        RecentAdd = Null_String,
        basebf    = Null_String;

    RecentAdd = buffer_filename ();
    %   file associated with the buffer just saved

    llen = strlen(RecentAdd);
    ifnot(llen) return;

    if (RecentAdd == LAST_SAVED) return;
    %   when you edit one file and save it again and again...

    if (RecentAdd == RecentFiles) return;
    if (RecentAdd == RecentDirs) return;
    %   disregard the two lists of recent files and dirs
    %   they cannot be added to themselves
    %   they would segfault because of the hook

    %   disregard backups terminating with tilde
    if ("~" == substr(RecentAdd,llen,1)) return;

    %   disregard tmp files
    if (is_substr(RecentAdd, ".tmp")) return;

    %   red is an associative array to indicate
    %   whether a file has already been added
    %   to the recent list in this editor session.
    %   if key exists, the same file can be
    %   saved and reloaded in the same session
    %   but only one entry will go for it to the recent list

    if (assoc_key_exists (red, whatbuf)) return;
    else red[whatbuf]=1;

    %   we are not in the recent files list
    %   we are in a buffer that was just saved

    % track down random corruption of file
    variable tst = 0;
    tst = copy_file (RecentFiles, JED_ROOT+"/f" + string(_time));
    if(tst < 0)
        write_scratch (JED_ROOT+"/f" + string(_time));

    %   we pre-pend the file just saved to the recent files list
    variable tmpbf = whatbuf();
    () = find_file (RecentFiles);
    variable tbf = whatbuf();
    bob();
    newline();
    insert(RecentAdd);
    newline();
    () = write_buffer (RecentFiles);
    flush(" ");
    sw2buf(tmpbf);
    delbuf (tbf);

    LAST_SAVED = RecentAdd;
    %   The file just saved tops the recent files list
}
add_to_hook ("_jed_save_buffer_after_hooks", &recent__files_update);

define recent__dirs_update ()
%----------------------------------------------------------
%   Will add the current directory to the list of recently
%   accessed dirs. Update is dependent on opening an existing
%   disk file and so accessing a directory. Access of same
%   directory is recorded only once per session.
%----------------------------------------------------------
{
    variable fst = file_status (RecentDirs);
    if (fst < 0) return;
    if (fst > 1) return;

    variable
        ddir = Null_String,
        bf   = Null_String,
        bn   = Null_String;

    %   current buffer holds the file just opened

    (bf, ddir, bn,) = getbuf_info ();

    if (ddir == LASTDIR) return;
    %   when you open files from one directory again and again...

    if (bf == Null_String)          return;
    if (is_substr(bn, slash))       return;
    if (is_substr(RecentFiles,bn))  return;
    if (is_substr(RecentDirs,bn))   return;

    if (file_status (ddir) != 2)
    {
        delbuf (bf);
        flush("Invalid directory: " + ddir);
        return;
    }
    %   can happen when new buffer created
    %   with file not yet on disk

    %   r_dir is an associative array to indicate
    %   whether a dir has already been added
    %   to the recent list in this editor session.
    %   if key exists, no further entry for the dir
    %   will go to the recent list

    if (assoc_key_exists (r_dir, ddir)) return;
    else r_dir[ddir]=1;

    %   we are not in the recent dirs list
    %   we are in a buffer where a file just opned from disk

    % track down random corruption of file
    variable tst = 0;
    tst = copy_file (RecentDirs, JED_ROOT+"/d" + string(_time));
    if(tst < 0)
        write_scratch (JED_ROOT+"/d" + string(_time));

    %   we pre-pend the file just saved to the recent dirs list
    variable tmpbf = whatbuf();
    () = find_file (RecentDirs);
    variable tbf = whatbuf ();
    bob();
    newline();
    insert(ddir);
    newline();
    () = write_buffer (RecentDirs);
    flush(" ");
    sw2buf(tmpbf);
    delbuf(tbf);

    LASTDIR = ddir;
    %   Current buffer dir is at top of recent directories list
}
add_to_hook ("_jed_find_file_after_hooks", &recent__dirs_update);

define recent_up ()
{
    go_up(1);

    if(is_whiteline and what_line < 6)
        () = fsearch (slash);

    else if(is_whiteline)
        () = bsearch (slash);

    bol();
}

define recent_down ()
{
    go_down(1);

    if(is_whiteline and what_line < 6)
        () = fsearch (slash);

    else if(is_whiteline)
        () = bsearch (slash);

    bol();
}

define recent_keymap ()
%----------------------------------------------------------
%   defines the keymap for the list of recent files/dirs
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "recent_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<256; i++) undefinekey (char(i), $1);
    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Up,    $1);
    undefinekey (Key_Down,  $1);
    undefinekey (Key_Home,  $1);
    undefinekey (Key_End,   $1);
    undefinekey (Key_Del,   $1);
    undefinekey (Key_BS,    $1);
    undefinekey ("^S",      $1);

    definekey ("recent_quit",           "^W",       $1);
    definekey ("recent_remove_line",    Key_Del,    $1);
    definekey ("recent_down",           Key_Down,   $1);
    definekey ("recent_load",           Key_Enter,  $1);
    definekey ("recent_quit",           Key_Escape, $1);
    definekey ("recent_switch",         Key_Tab,    $1);
    definekey ("recent_up",             Key_Up,     $1);
}

define recent_files ()
%----------------------------------------------------------
%   List max. 32 recently saved files
%   and allow opening them.
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    if (1 !=  file_status (RecentFiles))
    {
        flush("No list of recently edited files.");
        return;
    }

    %   if the recent files list is being edited, close it
    variable basebf = path_basename(RecentFiles);
    if (bufferp(basebf)) delbuf (basebf);

    delbuf (RecentzBuf);
    sw2buf (RecentzBuf);
    set_buffer_no_autosave ();
    () = insert_file (RecentFiles);

    bob();
    while(fsearch(" /"))
    {
        bol();
        trim();
        eol();
    }

    bob();
    ifnot (bol_fsearch (slash))
    {
        delbuf(whatbuf);
        () = delete_file (RecentFiles);
        flush("No list of recently edited files.");
        recent_quit();
        return;
    }

    variable
        i=0,
        j=0,
        L=64,
        ffilez = String_Type [L];

    ffilez [*] = Null_String;

    %   load lines into array for further operations

    bob();
    while (bol_fsearch(slash))
    {
        if (i > L-1) break;
        push_mark ();
        eol();
        ffilez[i] = bufsubstr ();
        i++;
    }
    L=i;

    %   mark repeated entries (turn into Null_String)
    %   same for files previously edited but not avalable any longer
    for (i=0; i<L; i++)
    {
        if (ffilez[i]==Null_String) continue;

        if (is_substr(ffilez[i], "//"))
        {
            ffilez[i] = Null_String;
            continue;
        }

        if (file_status (ffilez[i]) != 1)
        {
            ffilez[i] = Null_String;
            continue;
        }

        for (j=i+1; j<L; j++)
        {
            if (ffilez[i] == ffilez[j]) ffilez[j] = Null_String;
        }
    }

    %   shrink the list, drop empty lines
    ffilez = ffilez[where(ffilez!=Null_String)];

    variable llen = length(ffilez);

    ifnot (llen)
    {
        delbuf(whatbuf);
        () = delete_file (RecentFiles);
        flush("No list of recently edited files.");
        recent_quit();
        return;
    }

    if (llen > max_r) llen = max_r;

    %   save edited list without triggering hooks
    %   first comes the new entry, then the old entries
    %   in the original order

    () = delete_file(RecentFiles);
    variable fp = fopen (RecentFiles, "a");
    for (i=0; i < llen ; i++)
        () = fputs (ffilez[i]+char(10), fp);
    () = fclose (fp);

    %   now create pick list

    erase_buffer(whatbuf);
    loop(3) newline();

    go_up(2);
    goto_column (5);
    insert ("Press Enter on highlighted file to open it.");
    newline ();
    goto_column (5);
    insert ("[ Shift-Enter to open path only. ]");
    loop(3) newline();

    %   indent file names

    for (i=0; i < llen ; i++)
    {
        insert ("    ");
        insert (ffilez[i]);
        newline();
    }
    eob();
    loop(20) newline();
    %   so the end-of-buffer tag is not seen

    %   recent list is in special buffer

    bob();
    () = fsearch(slash);
    recent_keymap();

    set_buffer_modified_flag (0);
    set_readonly (1);

    set_status_line
    (" Escape: exit  | Up, Down: navigate  | Del: remove  | Enter: open  | Tab: dirs", 0);

    %   hide the editing menu bar
    variable m = "&Recently Edited Files";
    dummy_menu(m);

    %   set globl_str to tell files list from directories list
    globl_str = m;
    flush(" ");
}

define recent_dirs ()
%----------------------------------------------------------
%   Will list max. 32 recently accessed directories
%   and allow opening them.
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    if (1 !=  file_status (RecentDirs))
    {
        flush("No list of recently accessed directories.");
        return;
    }

    %   if the recent dirs list is being edited, close it
    variable basebf = path_basename(RecentDirs);
    if (bufferp(basebf)) delbuf (basebf);

    delbuf (RecentzBuf);
    sw2buf (RecentzBuf);
    set_buffer_no_autosave ();
    () = insert_file (RecentDirs);

    bob();
    while(fsearch(" /"))
    {
        bol();
        trim();
        eol();
    }

    bob();
    ifnot (bol_fsearch (slash))
    {
        delbuf(whatbuf);
        () = delete_file(RecentDirs);
        flush("No list of recently accessed directories.");
        recent_quit();
        return;
    }

    variable
        i=0,
        j=0,
        L=64,
        ffilez = String_Type [L];

    ffilez [*] = Null_String;

    %   load entries into array for further operations
    bob();
    while (bol_fsearch(slash))
    {
        if (i > L-1) break;
        push_mark ();
        eol();
        ffilez[i] = bufsubstr ();
        i++;
    }
    L=i;

    %   mark repeated entries (turn into Null_String)
    %   same for entries not avalable any longer
    for (i=0; i<L; i++)
    {
        if (ffilez[i]==Null_String) continue;

        if (is_substr(ffilez[i], "//"))
        {
            ffilez[i] = Null_String;
            continue;
        }

        if (file_status (ffilez[i]) != 2)
        {
            ffilez[i] = Null_String;
            continue;
        }

        for (j=i+1; j<L; j++)
        {
            if (ffilez[i] == ffilez[j]) ffilez[j] = Null_String;
        }
    }

    %   shrink the list, drop empty lines
    ffilez = ffilez[where(ffilez!=Null_String)];

    variable llen = length(ffilez);

    ifnot (llen)
    {
        delbuf(whatbuf);
        () = delete_file(RecentDirs);
        flush("No list of recently accessed directories.");
        recent_quit();
        return;
    }

    if (llen > max_r) llen = max_r;

    %   save edited list without triggering hooks.
    %   first comes the new entry, then the old entries
    %   in the original order.

    () = delete_file(RecentDirs);
    variable fp = fopen (RecentDirs, "a");
    if (fp == NULL)
    {
        flush("Cannot write to " + RecentDirs);
        () = fclose (fp);
        return;
    }
    for (i=0; i < llen ; i++)
        () = fputs (ffilez[i] + char(10), fp);
    () = fclose (fp);

    %   now create pick list

    erase_buffer(whatbuf);
    loop(3) newline();

    go_up(2);
    goto_column (5);
    insert ("Press Enter on highlighted directory to open it.");
    loop(3) newline();

    %   indent file names

    for (i=0; i < llen ; i++)
    {
        insert ("    ");
        insert (ffilez[i]);
        newline();
    }
    eob();
    loop(20) newline();

    %   recent dirs in special buffer

    bob();
    () = fsearch(slash);
    recent_keymap();

    set_buffer_modified_flag (0);
    set_readonly (1);

    set_status_line
    (" Escape: exit  | Up, Down: navigate  | Del: remove  | Enter: open  | Tab: files", 0);

    %   hide the editing menu bar
    variable m = "&Recently Accessed Directories";
    dummy_menu(m);

    %   set globl_str to tell apart files and directories list
    globl_str = m;
    flush(" ");
}

define recent_switch ()
{
    if (globl_str=="&Recently Accessed Directories")
        recent_files();
    else
        recent_dirs();

    %   since there is no buffer switch this hook was failing
    set_buffer_hook ("update_hook", &shape_cursor1);
}

%**********************************************************
%   Manage the shell buffer
%**********************************************************

variable bf_shell = "*scratch*";

define shell_prompt ()
{
    variable dir;

    (,dir,,) = getbuf_info ();
    %   get directory

    sw2buf(bf_shell);
    %   only relevant when coming from a different buffer
    if(markp) toggle_block (0);

    eob();
    shell_exec ("cd "+dir);

    newline();
    insert("Type command to the shell, press  ^E  to execute.");
    newline();
    newline();
}

define shell ()
%----------------------------------------------------------
%   Will switch to *scratch* to send commands to the shell
%   and receive output from it
%----------------------------------------------------------
{
    if (whatbuf == bf_shell)
    {
         shell_input();
         return;
    }

    shell_prompt();

    %   switched, current buffer is *scratch*

    setbuf_info ((getbuf_info | 0x100) & ~(0xA3));
    %   set buffer flags as follows
    %   no backup       (0x100)
    %   no save         ~(0x80)
    %   no undo         ~(0x20)
    %   no autosave     ~(0x2)
    %   unmodified      ~(0x1)
    %   (all of them)   ~(0xA3)
}

define shell_exec (cmd)
%----------------------------------------------------------
%
%   Will forward 'cmd' to the OS shell except:
%
%   `cd [dir]'      change directory
%
%----------------------------------------------------------
{
    variable argv, cmd1 = strcompress (cmd, " \t");

    argv = extract_element (cmd, 0, ' ');
    %   parse cmd

    eob();
    newline();

    if (argv=="cd")
    {
        %   change directory
        argv = extract_element (cmd1, 1, ' ');
        if (argv == NULL) argv = "~/";
        if (argv == ".") argv = " ";
        if (argv == "..") argv = " ";
        argv = expand_filename (argv);

        if (change_default_dir (argv))
        {
            newline();
            insert ("Unable to change directory!");
            newline();
        }
        else
        {
            variable buf, flag;
            (,,buf,flag) = getbuf_info ();
            setbuf_info (Null_String, argv, buf, flag);
            insert("Current directory:");
            newline();
            newline();
            vinsert("    %s", argv);
            newline();
        }
        return "//";
        %   do not return null string, bug
    }

    else return cmd;
}

define shell_input ()
%----------------------------------------------------------
%   Ancillary function to shell ()
%----------------------------------------------------------
{
    if (markp)
    {
        flush (s_reminder);
        return;
    }

    variable cmd = Null_String;

    %   command is current line as string
    %   cursor jumps to end-of-buffer before executing

    bol ();
    push_mark ();
    eol ();
    cmd = strtrim(bufsubstr ());

    if (strlen(cmd))
    {
        push_mark ();
        eol ();
        if (markp) del_region ();
        newline();

        cmd = shell_exec (cmd);

        if (cmd != "//")
        {
            try
            {
                () = run_shell_cmd (cmd + " 2>&1");
            }
            catch AnyError:
            {
                flush("Unable to issue shell command.");
                delbuf(whatbuf);
                return;
            }
        }
    }

    shell_prompt ();
}

%----------------------------------------------------------
%   bookmarks
%
%   bMax   max.of number of bookmarks in a single buffer
%   bkm    array to associate buffer and bookmarks
%   refer  name of buffer taking the bookmarks
%   bm_K   cursor column in buffer when listing bookmarks
%   bm_L   cursor line   in buffer when listing bookmarks
%   bms    special buffer for listing bookmarks
%----------------------------------------------------------

variable
    bMax   = 10,
    refer  = Null_String,
    bm_K   = 0,
    bm_L   = 0,
    bms    = " /bookmarks/";

define bm_up ()
{
    go_up(1);

    if(is_whiteline and what_line < 6)
    {
        () = fsearch (":  ");
    }

    else if(is_whiteline)
        () = bsearch (":  ");

    bol();
}

define bm_down ()
{
    go_down(1);

    if(is_whiteline and what_line < 6)
    {
        () = fsearch (":  ");
    }

    else if(is_whiteline)
        () = bsearch (":  ");

    bol();
}

define bm_set()
{
    if (markp)
    {
        flush(s_reminder);
        return;
    }

    if(is_whiteline)
    {
        flush("No bookmarks on empty lines.");
        return;
    }

    variable tmp  = Mark_Type [bMax];
    %   for Mark_Type s. paste.c

    refer = whatbuf();

    if (assoc_key_exists (bkm, refer))
    {
        variable i, n;

        %   get the associated bookmarks array
        tmp = bkm [refer];

        %   avoid duplicates
        variable x = Mark_Type [1];
        x[0] = create_user_mark ();

        %   if this new mark is same as any of tmp elements
        %   just stop, nothing to do

        for (i=0; i < bMax; i++)
        {
            if (tmp[i]==x[0])
            {
                flush("Bookmark already set.");
                return;
            }
        }

        %   if control here, the new bookmark is really new
        %   any index available for it?

        for (i=0; i < bMax; i++)
        {
            if (tmp[i]==NULL)
            {
                n=i;
                %   the smallest one, if any
                break;
            }
        }

        if (i == bMax)
        {
            %   no free element available
            %   kick out bookmark with index 0
            %   other bookmarks get decreased index
            %   use top index for new bookmark

            for (i=0; i < bMax-1; i++)
                tmp[i] = tmp[i+1];

            n = bMax-1;
        }

        tmp[n] = x[0];
    }

    else
    {
        %   associate to buffer a bookmark array
        bkm [refer] = tmp;

        %   first element will take bookmark
        tmp[0] = create_user_mark();
    }

    flush("Bookmark set at cursor.");
}

define bm_line ()
{
    bol ();
    ifnot (ffind (":  ")) return;

    push_mark();
    () = bfind(" ");
    variable ln = integer (bufsubstr);

    %   we just search on same line for a string

    () = ffind (":  ");
    go_right(3);
    variable bm_str = bfss ();
    delbuf(bms);

    sw2buf(refer);
    goto_line (ln);
    () = ffind(bm_str);
}

define bm_cleanup()
{
    delbuf (bms);
    sw2buf(refer);
    assoc_delete_key (bkm, refer);
    goto_line(bm_L);
    goto_column(bm_K);
}

define bm_rank (U)
{
    %   U is a user mark

    if (U==NULL) return 0;

    goto_user_mark (U);
    return what_line();
}

define bm_list ()
{
    if (markp)
    {
        flush(s_reminder);
        return;
    }

    refer = whatbuf();
    bm_L = what_line();
    bm_K = what_column();

    ifnot (assoc_key_exists (bkm, refer))
    {
        flush("No bookmarks set.");
        return;
    }

    variable
        B     = Integer_Type [bMax],
        tmp   = Mark_Type [bMax],
        i     = 0,
        lnstr = String_Type [bMax],
        ln    = Integer_Type [bMax];

    tmp = bkm [refer];
    B   = array_map (Integer_Type, &bm_rank, tmp);
    B   = array_sort (B);
    tmp = tmp[B];

    for (i=0; i < bMax; i++)
    {
        if (tmp[i]==NULL)
        {
            lnstr[i] = Null_String;
            ln[i]    = 0;
            continue;
        }

        goto_user_mark (tmp[i]);

        % the following has the side effect to bury bookmarks at eob
        lnstr[i] = strtrim(bfss);
        ln[i] = what_line();
    }

    %   initialize list
    sw2buf(bms);
    bob();

    variable lndx = Null_String;

    for (i=0; i < bMax; i++)
    {
        ifnot(strlen(lnstr[i])) continue;

        lndx = strcat("  [", string(B[i]), "] ");
        insert(lndx);

        if (ln[i] < 100000)
            vinsert ("%5d:  ", ln[i]);

        else
            vinsert ("%5s:  ", "*****");

        insert (lnstr[i]);
        newline ();
    }

    push_spot ();
    ifnot ( bsearch (":  "))
    {
        pop_spot();
        flush("No bookmarks found.");
        bm_cleanup ();
        return;
    }
    pop_spot();

    newline();
    newline();
    newline();
    insert("  Up, Down:   navigate");
    newline();
    insert("  Delete:     remove current bookmark");
    newline();
    insert("  F5:         remove all bookmarks");

    bob();
    newline ();
    insert("  ");
    loop(60) insert("-");
    newline();
    insert("   #   ");
    insert("line");
    insert("   ");
    insert("string at bookmark");
    newline();
    insert("  ");
    loop(60) insert("-");
    newline();
    newline();

    bm_keymap();

    set_buffer_modified_flag (0);
    set_readonly (1);
    hide_status_indicators(1);

    set_status_line
    ("  Escape: exit   |   Enter: go to bookmark", 0);

    %   heading in menu bar
    variable m = "&Bookmarks in buffer: " + refer;
    dummy_menu(m);
    set_buffer_hook ("update_hook", &shape_cursor1);
}

define bm_prev ()
{
    if (markp)
    {
        flush(s_reminder);
        return;
    }

    refer = whatbuf();

    ifnot (assoc_key_exists (bkm, refer))
    {
        flush("No bookmarks set.");
        return;
    }

    variable
        B     = Integer_Type [bMax],
        tmp   = Mark_Type [bMax],
        i     = 0,
        ok    = 0,
        ln    = 0,
        lna   = what_line(),
        cna   = what_column();

    tmp = bkm [refer];
    B   = array_map (Integer_Type, &bm_rank, tmp);
    tmp = tmp[array_sort (B)];

    for (i=bMax-1; i > -1; i--)
    {
        if (tmp[i]==NULL)
            continue;

        goto_user_mark (tmp[i]);
        ln = what_line();

        if (ln > lna) continue;
        else if (ln < lna) ok=1;
        else if(what_column < cna) ok=1;

        if(ok) break;
    }

    %   sorting may fail for bookmarks on same line

    if(ok)
        flush("Cursor is on bookmark.");

    else
    {
        goto_line(lna);
        goto_column(cna);
        flush("No bookmarks ahead of cursor.");
    }
}

define bm_next ()
{
    if (markp)
    {
        flush(s_reminder);
        return 0;
    }

    refer = whatbuf();

    ifnot (assoc_key_exists (bkm, refer))
    {
        flush("No bookmarks set.");
        return 0;
    }

    variable
        B     = Integer_Type [bMax],
        tmp   = Mark_Type [bMax],
        i     = 0,
        ok    = 0,
        ln    = 0,
        lna   = what_line(),
        cna   = what_column();

    tmp = bkm [refer];
    B   = array_map (Integer_Type, &bm_rank, tmp);
    tmp = tmp[array_sort (B)];

    for (i=0; i < bMax; i++)
    {
        if (tmp[i]==NULL) continue;

        goto_user_mark (tmp[i]);
        if(eobp) continue;
        ln = what_line();

        if (ln < lna) continue;
        else if (ln > lna) ok=1;
        else if((what_column > cna) and not eolp) ok=1;

        if(ok) break;
    }

    %   sorting may fail when several bookmarks on same line

    if(ok)
    {
        flush("Cursor is on bookmark.");
        ok = 0;
    }

    else
    {
        goto_line(lna);
        goto_column(cna);
        flush("No bookmarks past cursor.");
        ok = -1;
    }

    return ok;
}

define bm_zap ()
%----------------------------------------------------------
%   move user mark to eob and let it disappear!
%----------------------------------------------------------
{
    bol();
    ifnot(ffind("[")) return;

    go_right(1);
    push_mark();
    go_right(1);
    variable ndx = integer(bufsubstr);
    variable tmp = Mark_Type [bMax];
    tmp = bkm [refer];

    set_readonly (0);
    delete_line ();
    set_readonly (1);

    if(is_whiteline)
    {
        go_up(1);
        bol();
    }

    ifnot(bol_fsearch("  ["))
    {
        bm_cleanup ();
        return;
    }

    %   bookmarks list is not empty
    %   now do zap that bookmark

    sw2buf(refer);
    eob();
    move_user_mark (tmp[ndx]);
    sw2buf(bms);
}

define bm_keymap()
{
    reset_slim_keymap ();
    $1 = "bm_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i<256; i++) undefinekey (char(i), $1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Home,  $1);
    undefinekey (Key_End,   $1);

    definekey ("slim_quit",     "^W",       $1);
    definekey ("bm_cleanup",    Key_F5,     $1);
    definekey ("bm_down",       Key_Down,   $1);
    definekey ("bm_line",       Key_Enter,  $1);
    definekey ("slim_quit",     Key_Escape, $1);
    definekey ("bm_up",         Key_Up,     $1);
    definekey ("bm_zap",        Key_Del,    $1);
}

%----------------------------------------------------------
%   search dialog
%----------------------------------------------------------

variable s_fline   = Int_Type [4];
variable s_optstr  = String_Type [3,2];
variable s_opt_no  = Int_Type [3];

variable s_buffr  ;
variable s_cline  ;
variable s_current;
variable s_dline  ;
variable s_icolumn;
variable s_iline  ;
variable s_tline  ;

define s_init ()      ;
define s_paste ()     ;
define s_delete ()    ;
define s_bkspace ()   ;
define s_right ()     ;
define s_left ()      ;
define s_2field ()    ;
define s_insert ()    ;
define s_option ()    ;
define s_quit ()      ;
define s_prev ()      ;
define s_next ()      ;
define s_go ()        ;
define s_keymap ()    ;


define s_srch ()
{
    if(markp) toggle_block (0);

    variable is_split = 0;
    is_split = nwindows() - MINIBUFFER_ACTIVE - 1;

    if(is_split)
    {
        search_forward ();
        %   split already in use, thus simple search, case-insensitive

        return;
    }

    variable fn = Null_String;
    variable spacez4 = "    ";

    s_buffr = whatbuf();
    fn = strcat ("tmp", string(_time), slash);
    delbuf(fn);
    pop2buf(fn);

    s_init();

    hide_status_indicators(1);

    set_status_line
    ("  Search Dialog   |   Escape: cancel   |   Enter: start search", 0);

    variable linez;
    variable n, h;
    n = 15;
    h = 1;
    linez = String_Type [n];
    linez[*] = " ";

    linez[h+2] =
    "* Type of Search    ?simple         (simple / regex)";

    linez[h+3] =
    "* Direction         ?forward        (forward / backward)";

    linez[h+4] =
    "* Case-sensitive    ?no             (no / yes)";

    linez[h+5] =
    "* Search for        ?";

    linez[h+8] =
    "Tab:     next field                Space:  change option";

    linez[h+9] =
    "Up/Down: (search string) history   ctrl-v: paste (on search string)";

    variable i;
    for (i = 0; i < n; i++)
    {
        goto_line(i);
        if(strlen(string(linez[i])))
        {
            insert(spacez4);
            insert(string(linez[i]));
        }
        newline();
    }

    loop(SCREEN_HEIGHT - n) newline();
    %   hide eob mark

    bob();

    i=0;
    while(fsearch ("?"))
    {
        del();
        s_fline[i] = what_line ();
        i++;
    }

    s_current = s_iline;
    s_icolumn = what_column ();
    s_keymap();

    set_buffer_hook ("update_hook", &shape_cursor1);
}


define s_init ()
{
    s_opt_no [*] = 0;

    %   strings must have length 9 for vertical alignment
    s_optstr [*,*] =
    [
        ["simple   ",  "regex    "],
        ["forward  ",  "backward "],
        ["no       ",  "yes      "]
    ];

    s_current = 0;
    s_icolumn = 0;
    s_tline   = 0;
    s_dline   = 1;
    s_cline   = 2;
    s_iline   = 3;

    %   following has to match the defaults in linez
    %   namely s_optstr [i,j]
    %   i is the field, j is the option here below
    %   index in array is current field
    %   value is option for that field
    s_opt_no [0] = 0;
    s_opt_no [1] = 0;
    s_opt_no [2] = 0;
}

define s_paste ()
{
    if (s_current < s_iline) return;

    variable thisbf = whatbuf();
    EXIT_BLOCK sw2buf(thisbf);

    %   paste must fit into input field
    variable pmaxlen;
    push_spot();
    eol();
    pmaxlen = SCREEN_WIDTH - 2 - what_column();
    pop_spot();
    if(pmaxlen < 1) return;

    sw2buf("*scratch*");
    eob();
    newline();
    push_spot();
    call("yank");
    pop_spot();

    %   no need of white space
    trim();
    eol();
    trim();
    bol();
    if(eolp) return;

    eol();
    if (what_column > pmaxlen)
        goto_column(pmaxlen);

    push_mark();
    bol();
    call ("copy_region");
    %    mark popped

    push_mark();
    eob();
    del_region ();

    sw2buf(thisbf);
    call ("yank");
}


define s_delete ()
{
    if (s_current < s_iline) return;
    ifnot(eolp) del();
}


define s_bkspace ()
{
    if (s_current < s_iline) return;
    else if (what_column > s_icolumn)
        call ("backward_delete_char");
}


define s_right ()
{
    if (s_current < s_iline) return;
    else if ((what_column < SCREEN_WIDTH - 2) and not eolp)
         go_right(1);
}


define s_left ()
{
    if (s_current < s_iline);
    else if (what_column > s_icolumn)
        go_left(1);
}


define s_2field ()
{
    s_current++;
    s_current = s_current mod (s_iline + 1);

    % cursor moves down or cycles

    goto_line (s_fline[s_current]);
    goto_column (s_icolumn);
}


define s_insert ()
{
    if (s_current < s_iline) return;
    if (what_column < SCREEN_WIDTH - 2)
        call("self_insert_cmd");
}


define s_option ()
{
    if (s_current == s_iline)
    {
        s_insert();
        return;
    }

    if (s_current == s_cline)
    {
        if(s_opt_no[s_tline] == 0)
        {
            flush("Simple search can only be case-insensitive.");
            return;
        }
    }

    variable i,j, k,l;
    i = s_current;
    j = not (s_opt_no[s_current]);
    s_opt_no[s_current] = j;

    push_spot();
    if (s_current == s_tline)
    {
        % search type changed to simple?
        if(s_opt_no[s_tline] == 0)
        {
            %   case
            if(s_opt_no[s_cline] == 1)
            {
                s_opt_no[s_cline] = 0;
                k = s_opt_no[s_cline];
                l = s_cline;
                goto_line(s_fline[l]);
                goto_column(s_icolumn);

                loop(9) del();
                insert(s_optstr[l,k]);
                flush("Simple search can only be case-insensitive.");
            }
        }
    }
    pop_spot();

    push_spot();
    loop(9) del();
    insert(s_optstr[i,j]);
    pop_spot();
}

define s_quit ()
{
    variable tmpbf = whatbuf;
    toggle_windows();
    delbuf(tmpbf);
    del_tmpbuf();
}

define s_prev ()
%----------------------------------------------------------
%   Get previous command from minibuffer's history
%   'previous' means increased index, cycle around at top
%----------------------------------------------------------
{
    if (s_current != s_iline) return;

    goto_column(s_icolumn);
    push_mark ();
    eol ();
    del_region();
    mS = 0;

    if (strlen(Mini_History[mS, Mini_Last[mS]]) < heml)
    {
        Mini_Last[mS]=0;

        if (strlen(Mini_History[mS, Mini_Last[mS]]) < heml)
            return;
    }

    insert (Mini_History[mS, Mini_Last[mS]]);
    Mini_Last[mS]++;
    Mini_Last[mS] = Mini_Last[mS] mod Mini_Top;
}

define s_next ()
%----------------------------------------------------------
%   Get next command from minibuffer's history
%   'next' means decreased index, cycle around at bottom
%----------------------------------------------------------
{
    if (s_current != s_iline) return;

    goto_column(s_icolumn);
    push_mark ();
    eol ();
    del_region();
    mS = 0;

    variable ndx = Mini_Last[mS]-1;
    if (ndx < 0) ndx += Mini_Top;

    while (strlen(Mini_History[mS, ndx]) < heml and ndx > 0)
        ndx--;

    if (strlen(Mini_History[mS, ndx]) > heml-1)
        insert (Mini_History[mS,ndx]);

    Mini_Last[mS] = ndx;
}

define s_go ()
{
    variable sstr = Null_String;

    goto_line(s_fline[s_iline]);
    goto_column(s_icolumn);
    push_mark();
    eol();
    sstr = bufsubstr();

    ifnot(strlen(sstr))
    {
        flush("Search string/pattern missing.");
        return;
    }

    variable D;

    D = not s_opt_no [s_dline];
    % D=1 forward

    SEARCH_TYPE = s_opt_no [s_tline];

    CASE_SEARCH = s_opt_no [s_cline];
    % 1 sensitive

    s_quit ();
    sw2buf(s_buffr);

    mS = 0;
    Mini_Last[mS] = 0;
    mini_storeline (sstr);
    searchx (D, sstr);
}

define s_keymap()
{
    reset_slim_keymap ();
    $1 = "s_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i < 32; i++) undefinekey (char(i), $1);
    for (i=32; i < 127; i++)
        definekey ("s_insert", char(i), $1);
    for (i=160; i < 512; i++)
       definekey ("s_insert", char(i), $1);

    undefinekey (Key_Up,    $1);
    undefinekey (Key_Down,  $1);
    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Home,  $1);
    undefinekey (Key_End,   $1);
    undefinekey (Key_BS,    $1);

    definekey ("s_paste",   "^V",       $1);
    definekey ("s_quit",    "^W",       $1);
    definekey ("s_option",  Key_Space,  $1);
    definekey ("s_left",    Key_Left,   $1);
    definekey ("s_right",   Key_Right,  $1);
    definekey ("s_2field",  Key_Tab,    $1);
    definekey ("s_quit",    Key_Escape, $1);
    definekey ("s_delete",  Key_Del,    $1);
    definekey ("s_bkspace", Key_BS,     $1);
    definekey ("s_go",      Key_Enter,  $1);
    definekey ("s_prev",    Key_Up,     $1);
    definekey ("s_next",    Key_Down,   $1);
}

%----------------------------------------------------------
%   Occurrences:  dialog for options and string
%----------------------------------------------------------

variable zo_fline   = Int_Type [3];
variable zo_optstr  = String_Type [2,2];
variable zo_opt_no  = Int_Type [2];
variable zo_buffr   ;
variable zo_cline   ;
variable zo_current ;
variable zo_icolumn ;
variable zo_iline   ;
variable zo_tline   ;

define zo_2field()  ;
define zo_init()    ;
define zo_keymap()  ;
define zo_paste()   ;

define o_srch ()
{
    %   occurrences simple/regex

    variable ll = markp();

    %   use string selected on one line, if any
    if(ll==1 and what_block==1) copy_rect();
    else if(ll==1) call("copy_region");

    variable fn = Null_String;
    variable spacez4 = "    ";

    %   new buffer
    zo_buffr = whatbuf();
    fn = strcat ("tmp", string(_time), slash);
    delbuf(fn);
    pop2buf(fn);

    zo_init();
    hide_status_indicators(1);

    set_status_line
    ("  Occurrences Dialog   |   Escape: cancel   |   Enter: search", 0);

    variable linez;
    variable n, h;
    n = 15;
    h = 1;
    linez = String_Type [n];
    linez[*] = " ";

    linez[h+1] =
    "* Type of Search    ?simple      (simple / regex)";

    linez[h+2] =
    "* Case-sensitive    ?no          (no / yes)";

    linez[h+3] =
    "* String/Pattern    ?";

    linez[h+5] =
    "Tab:     next field                Space:  change option";

    linez[h+6] =
    "Up/Down: (search string) history   ctrl-v: paste (on search string)";

    if(ll > 1)
    {
        linez[h+8] = "Search will loose the current selection.";
        linez[h+9] = "Occurrences in selection will be tagged.";
    }

    variable i;
    for (i = 0; i < n; i++)
    {
        goto_line(i);
        if(strlen(string(linez[i])))
        {
            insert(spacez4);
            insert(string(linez[i]));
        }
        newline();
    }

    loop(SCREEN_HEIGHT - n - 2) newline();
    %   hide eob mark

    bob();

    i=0;
    while(fsearch ("?"))
    {
        del();
        zo_fline[i] = what_line ();
        i++;
    }

    zo_current = zo_iline;
    zo_icolumn = what_column ();
    if(ll==1)
    {
        zo_paste();
        goto_column(zo_icolumn);
    }

    zo_keymap();

    set_buffer_hook ("update_hook", &shape_cursor1);
}

define zo_init ()
{
    zo_current = 0;
    zo_icolumn = 0;
    zo_tline   = 0;
    zo_cline   = 1;
    zo_iline   = 2;

    zo_opt_no [*] = 0;

    %   strings must have length 9 for alignment with heading
    zo_optstr [*,*] =
    [
        ["simple   ", "regex    "],
        ["no       ", "yes      "]
    ];

    %   following has to match the defaults in linez
    %   namely zo_optstr [i,j]
    %   i is the field, j is the option here below
    %   index in array is current field
    %   value is option for that field
    zo_opt_no [0] = 0;
    zo_opt_no [1] = 0;
}

define zo_2field()
{
    zo_current++;
    zo_current = zo_current mod (zo_iline + 1);

    % cursor moves down or cycles

    goto_line (zo_fline[zo_current]);
    goto_column (zo_icolumn);
}

define zo_insert()
{
    if (zo_current < zo_iline) return;
    if (what_column < SCREEN_WIDTH - 2)
        call("self_insert_cmd");
}

define zo_option()
{
    if (zo_current == zo_iline)
    {
        zo_insert();
        return;
    }

    if (zo_current == zo_cline)
    {
        if(zo_opt_no[zo_tline] == 0)
        {
            flush("Simple search can only be case-insensitive.");
            return;
        }
    }

    variable i,j, k,l;
    i = zo_current;
    j = not (zo_opt_no[zo_current]);
    zo_opt_no[zo_current] = j;

    push_spot();
    if (zo_current == zo_tline)
    {
        % search type changing to simple
        if(zo_opt_no[zo_tline] == 0)
        {
            %   case
            if(zo_opt_no[zo_cline] == 1)
            {
                zo_opt_no[zo_cline] = 0;
                k = zo_opt_no[zo_cline];
                l = zo_cline;
                goto_line(zo_fline[l]);
                goto_column(zo_icolumn);

                loop(9) del();
                insert(zo_optstr[l,k]);
                flush("Simple search can only be case-insensitive.");
            }
        }
    }
    pop_spot();

    push_spot();
    loop(9) del();
    insert(zo_optstr[i,j]);
    pop_spot();
}

define zo_bkspace()
{
    if (zo_current < zo_iline) return;
    else if (what_column > zo_icolumn)
        call ("backward_delete_char");
}

define zo_delete ()
{
    if (zo_current < zo_iline) return;
    ifnot(eolp) del();
}

define zo_right()
{
    if (zo_current < zo_iline) return;
    else if ((what_column < SCREEN_WIDTH - 2) and not eolp)
         go_right(1);
}

define zo_left()
{
    if (zo_current < zo_iline);
    else if (what_column > zo_icolumn)
        go_left(1);
}

define zo_paste ()
{
    if (zo_current < zo_iline) return;

    variable thisbf = whatbuf();
    EXIT_BLOCK sw2buf(thisbf);

    %   paste must fit into input field
    variable pmaxlen;
    push_spot();
    eol();
    pmaxlen = SCREEN_WIDTH - 2 - what_column();
    pop_spot();
    if(pmaxlen < 1) return;

    sw2buf("*scratch*");
    eob();
    newline();
    push_spot();
    call("yank");
    pop_spot();

    %   no need of white space
    trim();
    eol();
    trim();
    bol();
    if(eolp) return;

    eol();
    if (what_column > pmaxlen)
        goto_column(pmaxlen);

    push_mark();
    bol();
    call ("copy_region");
    %    mark popped

    push_mark();
    eob();
    del_region ();

    sw2buf(thisbf);
    call ("yank");
}

define zo_prev ()
%----------------------------------------------------------
%   Get previous command from minibuffer's history
%   'previous' means increased index, cycle around at top
%----------------------------------------------------------
{
    if (zo_current != zo_iline) return;

    goto_column(zo_icolumn);
    push_mark ();
    eol ();
    del_region();
    mS = 0;

    if (strlen(Mini_History[mS, Mini_Last[mS]]) < heml)
    {
        Mini_Last[mS]=0;

        if (strlen(Mini_History[mS, Mini_Last[mS]]) < heml)
            return;
    }

    insert (Mini_History[mS, Mini_Last[mS]]);
    Mini_Last[mS]++;
    Mini_Last[mS] = Mini_Last[mS] mod Mini_Top;
}

define zo_next ()
%----------------------------------------------------------
%   Get next command from minibuffer's history
%   'next' means decreased index, cycle around at bottom
%----------------------------------------------------------
{
    if (zo_current != zo_iline) return;

    goto_column(zo_icolumn);
    push_mark ();
    eol ();
    del_region();
    mS = 0;

    variable ndx = Mini_Last[mS]-1;
    if (ndx < 0) ndx += Mini_Top;

    while (strlen(Mini_History[mS, ndx]) < heml and ndx > 0)
        ndx--;

    if (strlen(Mini_History[mS, ndx]) > heml-1)
        insert (Mini_History[mS,ndx]);

    Mini_Last[mS] = ndx;
}

define zo_quit()
{
    variable tmpbf = whatbuf;
    toggle_windows();
    delbuf(tmpbf);
    del_tmpbuf();
}

define zo_go()
{
    variable sstr = Null_String;

    goto_line(zo_fline[zo_iline]);
    goto_column(zo_icolumn);
    push_mark();
    eol();
    sstr = bufsubstr();

    ifnot(strlen(sstr))
    {
        flush("Search string missing.");
        return;
    }

    SEARCH_TYPE = zo_opt_no [zo_tline];

    CASE_SEARCH = zo_opt_no [zo_cline];
    % 1 sensitive

    zo_quit ();
    sw2buf(zo_buffr);

    occur(sstr);
}


define zo_keymap()
{
    reset_slim_keymap ();
    $1 = "zo_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i < 32; i++) undefinekey (char(i), $1);
    for (i=32; i < 127; i++)
        definekey ("zo_insert", char(i), $1);
    for (i=160; i < 512; i++)
       definekey ("zo_insert", char(i), $1);

    undefinekey (Key_Up,    $1);
    undefinekey (Key_Down,  $1);
    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Home,  $1);
    undefinekey (Key_End,   $1);
    undefinekey (Key_BS,    $1);

    definekey ("zo_paste",   "^V",       $1);
    definekey ("zo_quit",    "^W",       $1);
    definekey ("zo_option",  Key_Space,  $1);
    definekey ("zo_left",    Key_Left,   $1);
    definekey ("zo_right",   Key_Right,  $1);
    definekey ("zo_2field",  Key_Tab,    $1);
    definekey ("zo_quit",    Key_Escape, $1);
    definekey ("zo_delete",  Key_Del,    $1);
    definekey ("zo_bkspace", Key_BS,     $1);
    definekey ("zo_go",      Key_Enter,  $1);
    definekey ("zo_prev",    Key_Up,     $1);
    definekey ("zo_next",    Key_Down,   $1);
}


%----------------------------------------------------------
%   dialog for sorting
%----------------------------------------------------------

define ts_keymap();
define ts_init();
variable ts_optstr  = String_Type [2,2];
variable ts_opts_no = Integer_Type [2];
variable ts_current = 0;
variable ts_ocolumn = 0;
variable ts_oline   = 0;
variable ts_cline   = 1;
variable ts_fline   = Integer_Type [2];
variable ts_buffr   = Null_String;

define t_sort ()
%----------------------------------------------------------
%   start dialog to set sort options
%----------------------------------------------------------
{
    if (is_readonly) return;

    ifnot (markp)
    {
        flush("Selection missing: sorting requires a selection.");
        return;
    }

    if(markp==1)
    {
        flush("Nothing to do: only one line to sort.");
        return;
    }

    variable ts_mline;

    if(what_block==1)
        ts_mline = " Sort with fixed key length   ";
    else
        ts_mline = " Sort with variable key length   ";

    variable fn = Null_String;
    variable spacez4 = "    ";

    ts_buffr = whatbuf();

    %   new buffer
    fn = strcat ("tmp", string(_time), slash);
    delbuf(fn);
    pop2buf(fn);

    ts_init();
    hide_status_indicators(1);

    set_status_line
    (ts_mline + "|   Escape: cancel   |   Enter: sort", 0);

    variable linez;
    variable n, h;
    n = 15;
    h = 1;
    linez = String_Type [n];
    linez[*] = " ";

    linez[h+1] =
    "* Order              ?ascending    (ascending / descending)";

    linez[h+2] =
    "* Case-sensitive     ?no           (no / yes)";

    linez[h+5] =
    "Tab:    next field";

    linez[h+6] =
    "Space:  change option";

    variable i;
    for (i = 0; i < n; i++)
    {
        goto_line(i);
        if(strlen(string(linez[i])))
        {
            insert(spacez4);
            insert(string(linez[i]));
        }
        newline();
    }

    loop(SCREEN_HEIGHT - n) newline();
    %   hide eob mark

    bob();
    newline();

    for (i=0; i < 2; i++)
    {
        if(fsearch ("?"))
        {
            del();
            ts_fline[i] = what_line;
        }
    }

    ts_ocolumn = what_column ();
    ts_current = 0;

    goto_line(ts_fline[ts_oline]);
    goto_column(ts_ocolumn);

    %   following has to match the defaults in linez
    %   namely s_optstr [i,j]
    %   i is the field, j is the option here below
    %   index in array is current field
    %   value is option for that field
    ts_opts_no [0] = 0;
    ts_opts_no [1] = 0;

    ts_keymap();
    set_readonly (1);
    set_buffer_hook ("update_hook", &shape_cursor1);
    update (1);
}

define ts_option()
{
    variable i,j;
    i = ts_current;
    j = not (ts_opts_no[ts_current]);
    ts_opts_no[ts_current] = j;

    set_readonly (0);
    goto_column(ts_ocolumn);
    loop(10) del();
    insert(ts_optstr[i,j]);
    goto_column(ts_ocolumn);
    set_readonly (1);
}

define ts_init ()
{
    ts_opts_no [*] = 0;

    %   strings have length 10, some are padded
    ts_optstr [*,*] =
    [
        ["ascending ",  "descending"],
        ["no        ",  "yes       "]
    ];
}

define ts_2field()
{
    ts_current++;
    ts_current = ts_current mod 2;

    % cursor moves down or cycles

    goto_line (ts_fline[ts_current]);
    goto_column (ts_ocolumn);
}

define ts_quit()
{
    variable tmpbf = whatbuf;
    toggle_windows();
    delbuf(tmpbf);
    del_tmpbuf();
}

define ts_go()
{
    variable sstr = Null_String;

    variable D, E;

    D = ts_opts_no [ts_oline];
    % 0 ascending

    E = ts_opts_no [ts_cline];
    % 1 case-sensitive

    ts_quit();
    sw2buf(ts_buffr);

    if(what_block == 1) sort3 (E, D);
    else
    {
        sort4 (E, D);
        stream_to_lines();
    }
    %   cursor on same line as started
}

define ts_keymap()
{
    reset_slim_keymap ();
    $1 = "ts_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i<32; i++) undefinekey (char(i), $1);

    undefinekey (Key_Up,    $1);
    undefinekey (Key_Down,  $1);
    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Home,  $1);
    undefinekey (Key_End,   $1);
    undefinekey (Key_BS,    $1);

    definekey ("ts_quit",    "^W",       $1);
    definekey ("ts_option",  Key_Space,  $1);
    definekey ("ts_2field",  Key_Tab,    $1);
    definekey ("ts_quit",    Key_Escape, $1);
    definekey ("ts_go",      Key_Enter,  $1);
}


%----------------------------------------------------------
%   replace dialog
%----------------------------------------------------------

variable rpl_fline   = Int_Type [4];
variable rpl_optstr  = String_Type [2,2];
variable rpl_opt_no  = Int_Type [2];
variable rpl_buffr  ;
variable rpl_cline  ;
variable rpl_current;
variable rpl_icolumn;
variable rpl_iline  ;
variable rpl_iwhat  ;
variable rpl_iwith  ;
variable rpl_tline  ;

define rpl_2field() ;
define rpl_init()   ;
define rpl_keymap() ;

define r_srch ()
{
    variable is_split = 0;
    is_split = nwindows() - MINIBUFFER_ACTIVE - 1;

    if(is_split)
    {
        replace_interactivly();
        %   split already in use, thus simple search, case-insensitive

        return;
    }

    variable fn = Null_String;
    variable spacez4 = "    ";

    %   new buffer
    rpl_buffr = whatbuf();
    fn = strcat ("tmp", string(_time), slash);
    delbuf(fn);
    pop2buf(fn);

    rpl_init();
    hide_status_indicators(1);

    set_status_line
    ("  Replace Dialog   |   Escape: cancel   |   Enter: start search/replace", 0);

    variable linez;
    variable n, h;
    n = 15;
    h = 1;
    linez = String_Type [n];
    linez[*] = " ";

    linez[h+1] =
    "* Type of Search    ?simple         (simple / regex)";

    linez[h+2] =
    "* Case-sensitive    ?no             (no / yes)";

    linez[h+3] =
    "* Replace what      ?^";

    linez[h+4] =
    "* Replace with      ?";

    linez[h+6] =
    "Tab:     next field                Space:  change option";

    linez[h+7] =
    "Up/Down: (search string) history   ctrl-v: paste (on search string)";

    variable i;
    for (i = 0; i < n; i++)
    {
        goto_line(i);
        if(strlen(string(linez[i])))
        {
            insert(spacez4);
            insert(string(linez[i]));
        }
        newline();
    }

    loop(SCREEN_HEIGHT - n) newline();
    %   hide eob mark

    bob();

    i=0;
    while(fsearch ("?"))
    {
        del();
        rpl_fline[i] = what_line ();
        i++;
    }

    ()= bsearch ("^");
    del();
    rpl_current = rpl_iwhat;
    rpl_icolumn = what_column ();

    goto_line(SCREEN_HEIGHT - 4);
    goto_line(rpl_fline[rpl_iwhat]);
    goto_column(rpl_icolumn);

    rpl_keymap();
    set_buffer_hook ("update_hook", &shape_cursor1);
}


define rpl_init ()
{
    rpl_opt_no [*] = 0;

    %   strings must have length 9 for vertical alignment
    rpl_optstr [*,*] =
    [
        ["simple   ",  "regex    "],
        ["no       ",  "yes      "]
    ];

    rpl_current = 0;
    rpl_icolumn = 0;
    rpl_tline   = 0;
    rpl_cline   = 1;
    rpl_iwhat   = 2;
    rpl_iwith   = 3;

    %   following has to match the defaults in linez
    %   namely rpl_optstr [i,j]
    %   i is the field, j is the option here below
    %   index in array is current field
    %   value is option for that field
    rpl_opt_no [0] = 0;
    rpl_opt_no [1] = 0;
}

define rpl_paste ()
{
    ifnot (rpl_current > rpl_cline) return;

    variable thisbf = whatbuf();
    EXIT_BLOCK sw2buf(thisbf);

    %   paste must fit into input field
    variable pmaxlen;
    push_spot();
    eol();
    pmaxlen = SCREEN_WIDTH - 2 - what_column();
    pop_spot();
    if(pmaxlen < 1) return;

    sw2buf("*scratch*");
    eob();
    newline();
    push_spot();
    call("yank");
    pop_spot();

    %   no need of white space
    trim();
    eol();
    trim();
    bol();
    if(eolp) return;

    eol();
    if (what_column > pmaxlen)
        goto_column(pmaxlen);

    push_mark();
    bol();
    call ("copy_region");
    %    mark popped

    push_mark();
    eob();
    del_region ();

    sw2buf(thisbf);
    call ("yank");
}

define rpl_delete ()
{
    ifnot (rpl_current > rpl_cline) return;
    ifnot(eolp) del();
}

define rpl_bkspace()
{
    ifnot (rpl_current > rpl_cline) return;
    else if (what_column > rpl_icolumn)
        call ("backward_delete_char");
}

define rpl_right()
{
    ifnot (rpl_current > rpl_cline) return;
    else if ((what_column < SCREEN_WIDTH - 2) and not eolp)
         go_right(1);
}

define rpl_left()
{
    ifnot (rpl_current > rpl_cline);
    else if (what_column > rpl_icolumn)
        go_left(1);
}

define rpl_2field()
{
    rpl_current++;
    rpl_current = rpl_current mod (rpl_iwith + 1);

    % cursor moves down or cycles

    goto_line (rpl_fline[rpl_current]);
    goto_column (rpl_icolumn);
}

define rpl_insert()
{
    ifnot (rpl_current > rpl_cline) return;
    if (what_column < SCREEN_WIDTH - 2)
        call("self_insert_cmd");
}

define rpl_option()
{
    if (rpl_current > rpl_cline)
    {
        rpl_insert();
        return;
    }

    if (rpl_current == rpl_cline)
    {
        if(rpl_opt_no[rpl_tline] == 0)
        {
            flush("Simple search can only be case-insensitive.");
            return;
        }
    }

    variable i,j, k,l;
    i = rpl_current;
    j = not (rpl_opt_no[rpl_current]);
    rpl_opt_no[rpl_current] = j;

    push_spot();
    if (rpl_current == rpl_tline)
    {
        % search type changed to simple?
        if(rpl_opt_no[rpl_tline] == 0)
        {
            %   case
            if(rpl_opt_no[rpl_cline] == 1)
            {
                rpl_opt_no[rpl_cline] = 0;
                k = rpl_opt_no[rpl_cline];
                l = rpl_cline;
                goto_line(rpl_fline[l]);
                goto_column(rpl_icolumn);

                loop(9) del();
                insert(rpl_optstr[l,k]);
                flush("Simple search can only be case-insensitive.");
            }
        }
    }
    pop_spot();

    push_spot();
    loop(9) del();
    insert(rpl_optstr[i,j]);
    pop_spot();
}

define rpl_quit()
{
    variable tmpbf = whatbuf;
    toggle_windows();
    delbuf(tmpbf);
    del_tmpbuf();
}

define rpl_prev ()
%----------------------------------------------------------
%   Get previous command from minibuffer's history
%   'previous' means increased index, cycle around at top
%----------------------------------------------------------
{
    ifnot (rpl_current > rpl_cline) return;

    goto_column(rpl_icolumn);
    push_mark ();
    eol ();
    del_region();
    mS = 0;

    if (strlen(Mini_History[mS, Mini_Last[mS]]) < heml)
    {
        Mini_Last[mS]=0;

        if (strlen(Mini_History[mS, Mini_Last[mS]]) < heml)
            return;
    }

    insert (Mini_History[mS, Mini_Last[mS]]);
    Mini_Last[mS]++;
    Mini_Last[mS] = Mini_Last[mS] mod Mini_Top;
}

define rpl_next ()
%----------------------------------------------------------
%   Get next command from minibuffer's history
%   'next' means decreased index, cycle around at bottom
%----------------------------------------------------------
{
    ifnot (rpl_current > rpl_cline) return;

    goto_column(rpl_icolumn);
    push_mark ();
    eol ();
    del_region();
    mS = 0;

    variable ndx = Mini_Last[mS]-1;
    if (ndx < 0) ndx += Mini_Top;

    while (strlen(Mini_History[mS, ndx]) < heml and ndx > 0)
        ndx--;

    if (strlen(Mini_History[mS, ndx]) > heml-1)
        insert (Mini_History[mS,ndx]);

    Mini_Last[mS] = ndx;
}

define rpl_go()
{
    variable sstr = Null_String;
    variable rstr = Null_String;

    goto_line(rpl_fline[rpl_iwhat]);
    goto_column(rpl_icolumn);
    push_mark();
    eol();
    sstr = bufsubstr();

    ifnot(strlen(sstr))
    {
        flush("Search string missing, nothing to replace.");
        return;
    }

    % replacement can be empty
    goto_line(rpl_fline[rpl_iwith]);
    goto_column(rpl_icolumn);
    push_mark();
    eol();
    rstr = bufsubstr();

    SEARCH_TYPE = rpl_opt_no [rpl_tline];
    % 1 regex

    CASE_SEARCH = rpl_opt_no [rpl_cline];
    % 1 sensitive

    rpl_quit ();
    sw2buf(rpl_buffr);

    mS = 0;
    Mini_Last[mS] = 0;
    mini_storeline (sstr);


    %   switch to traditional interactive replacements
    replace_on_demand(sstr, rstr);
}

define rpl_keymap()
{
    reset_slim_keymap ();
    $1 = "rpl_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i < 32; i++) undefinekey (char(i), $1);
    for (i=32; i < 127; i++)
        definekey ("rpl_insert", char(i), $1);
    for (i=160; i < 512; i++)
       definekey ("rpl_insert", char(i), $1);

    undefinekey (Key_Up,    $1);
    undefinekey (Key_Down,  $1);
    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Home,  $1);
    undefinekey (Key_End,   $1);
    undefinekey (Key_BS,    $1);

    definekey ("rpl_paste",   "^V",       $1);
    definekey ("rpl_quit",    "^W",       $1);
    definekey ("rpl_option",  Key_Space,  $1);
    definekey ("rpl_left",    Key_Left,   $1);
    definekey ("rpl_right",   Key_Right,  $1);
    definekey ("rpl_2field",  Key_Tab,    $1);
    definekey ("rpl_quit",    Key_Escape, $1);
    definekey ("rpl_delete",  Key_Del,    $1);
    definekey ("rpl_bkspace", Key_BS,     $1);
    definekey ("rpl_go",      Key_Enter,  $1);
    definekey ("rpl_prev",    Key_Up,     $1);
    definekey ("rpl_next",    Key_Down,   $1);
}

%---------------------------------------------------
